"""
..currentmodule:: basin3d.core.sqlalchemy_models

:platform: Unix, Mac
:synopsis: SQLAlchemy models for BASIN-3D data sources
:module author: Valerie C. Hendrix <vchendrix@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
from typing import Any

from sqlalchemy import Column, Integer, String, ForeignKey, JSON, UniqueConstraint, Index
from sqlalchemy import create_engine
from sqlalchemy.types import Text
from sqlalchemy.orm import relationship, sessionmaker, declarative_base

Base: Any = declarative_base()
"""The base class for all SQLAlchemy models"""


class DataSource(Base):
    """
    Data source model for BASIN-3D data sources

    Fields:
        - *id:* autoincrement integer, primary key
        - *plugin_id:* string, unique identifier for the data source
        - *name:* string, unique name for the data source
        - *id_prefix:* string, prefix that is added to all data source ids
        - *location:* string, location of the data source (e.g., URL)
        - *plugin_module:* string, module name of the data source plugin
        - *plugin_class:* string, class name of the data source plugin
    """
    __tablename__ = 'data_source'
    id = Column(Integer, primary_key=True, autoincrement=True)
    plugin_id = Column(String(50), unique=True, nullable=False)
    name = Column(String(20), unique=True, nullable=False)
    id_prefix = Column(String(5), unique=True, nullable=False)
    location = Column(Text, nullable=True)
    plugin_module = Column(Text, nullable=True)
    plugin_class = Column(Text, nullable=True)

    __table_args__ = (
        Index('idx_id_prefix', 'id_prefix'),  # Adding an index on id_prefix
    )

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.name

    def __repr__(self):
        return '<DataSource %r>' % self.name


class ObservedProperty(Base):
    """
    Observed property model for BASIN-3D data sources

    Fields:
        - *basin3d_vocab:* string, unique identifier for the observed property
        - *full_name:* string, full name of the observed property
        - *categories:* string, delimited string of categories
        - *units:* string, units of the observed property
    """
    __tablename__ = 'observed_property'
    basin3d_vocab = Column(String(50), primary_key=True, nullable=False)
    full_name = Column(String(255), nullable=False)
    categories = Column(Text, nullable=True)  # Assuming categories are stored as a delimited string
    units = Column(String(50), nullable=False)

    __table_args__ = (
        Index('idx_basin3d_vocab', 'basin3d_vocab'),  # Adding an index on id_prefix
    )

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.description

    def __repr__(self):
        return '<ObservedProperty %r>' % self.basin3d_vocab


class AttributeMapping(Base):
    """
    Table to map attributes between BASIN-3D and data sources

    Fields:
        - *attr_type:* string, type of attribute
        - *basin3d_vocab:* string, unique identifier for the attribute
        - *basin3d_desc:* JSON, description of the attribute
        - *datasource_vocab:* string, unique identifier for the data source attribute
        - *datasource_desc:* text, description of the data source attribute
        - *datasource_id:* integer, foreign key to the data source

    """
    __tablename__ = 'attribute_mapping'
    id = Column(Integer, primary_key=True, autoincrement=True)
    attr_type = Column(String(50), nullable=False)
    basin3d_vocab = Column(String(50), nullable=False)
    basin3d_desc = Column(JSON, nullable=False)
    datasource_vocab = Column(String(50), nullable=False)
    datasource_desc = Column(Text, nullable=True)
    datasource_id = Column(Integer, ForeignKey('data_source.id'), nullable=False)
    datasource = relationship('DataSource')

    __table_args__ = (
        UniqueConstraint('datasource_id', 'attr_type', 'datasource_vocab', name='_datasource_attr_type_vocab_uc'),
        Index('idx_datasource_attr_type', 'datasource_id', 'attr_type', 'datasource_vocab'),
        # Adding an index on id_prefix
    )

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.datasource_vocab


def clear_database():
    """
    Clear the database
    """
    Base.metadata.drop_all(engine)
    Base.metadata.create_all(engine)


# Database setup in memory
# Create a temporary in memory SQLite database
engine = create_engine('sqlite:///:memory:')
Base.metadata.create_all(engine)
Session = sessionmaker(bind=engine)



"""

.. currentmodule:: basin3d.core.access

:platform: Unix, Mac
:synopsis: BASIN-3D ``DataSource`` access classes
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
from basin3d.core import monitor

logger = monitor.get_logger(__name__)


def get_url(url, params=None, headers=None, verify=True, **kwargs):
    """
    Send a GET request to the specified URL.  Note look up extra
    `kwargs` in `requests.get`
    :param url:
    :param params: request parameters
    :param headers: request headers
    :param verify: verify SSL connection
    :return: Response
    """
    import requests
    response = requests.get(url, params=params, verify=verify, headers=headers, **kwargs)
    logger.info("url:{}".format(response.url))
    return response


def post_url(url, params=None, headers=None, verify=True):
    """
    Send a POST request to the specified URL
    :param url:
    :param params: request parameters
    :param headers: request headers
    :param verify: verify SSL connection
    :return: Response
    """
    import requests
    response = requests.post(url, params=params, verify=verify, headers=headers)
    logger.info("url:{}".format(response.url))
    return response



"""

.. currentmodule:: basin3d.core.models

:synopsis: The BASIN-3D  Models
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""

import datetime
import enum
import json

from collections import namedtuple
from dataclasses import dataclass, field
from itertools import repeat
from numbers import Number
from typing import List, Optional, Union

from basin3d.core import monitor
from basin3d.core.schema.enum import FeatureTypeEnum, FEATURE_SHAPE_TYPES, MappedAttributeEnum, MAPPING_DELIMITER, NO_MAPPING_TEXT
from basin3d.core.translate import get_datasource_mapped_attribute, translate_attributes
from basin3d.core.types import SpatialSamplingShapes

logger = monitor.get_logger(__name__)


class JSONSerializable:
    """
    Make a Data class serializable to json
    """

    def to_json(self):
        def props(o):
            """Convert object to dict.  If prefixed with, _ remove it"""
            try:
                map = {}
                for k in o.__dict__.keys():
                    if not k.startswith("__"):
                        if k.startswith("_"):
                            map[k[1:]] = o.__dict__[k]
                        else:
                            map[k] = o.__dict__[k]
                return map
            except Exception:
                # There is not __dict__, return a string representation
                return str(o)

        return json.dumps(self, default=props,
                          sort_keys=True, indent=4)

    def to_dict(self):
        return json.loads(self.to_json())


@dataclass
class DataSource(JSONSerializable):
    """
    Data Source definition

    Fields:
        - *id:* string (inherited)
        - *name:* string
        - *id_prefix:* string, prefix that is added to all data source ids
        - *location:*
        - *credentials:*

    """
    id: str = ''  # not sure we need this id??
    name: str = ''
    id_prefix: str = ''
    location: str = ''
    credentials: dict = field(default_factory=dict)

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.name


@dataclass
class ObservedProperty(JSONSerializable):
    """
    Defining the properties being observed (measured). See http://vocabulary.odm2.org/variablename/ for controlled vocabulary

    Fields:
        - *basin3d_vocab:* string,
        - *full_name:* string,
        - *categories:* List of strings (in order of priority).
        - *units:* string

    See http://vocabulary.odm2.org/variabletype/ for options, although I think we should have our own list (theirs is a bit funky).
    """
    basin3d_vocab: str = ''
    full_name: str = ''
    categories: list = field(default_factory=list)
    units: str = field(default='')

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.basin3d_vocab


@dataclass
class AttributeMapping(JSONSerializable):
    """
    A data class for attribute mappings between datasource vocabularies and BASIN-3D vocabularies.
    These are the associations defined in the datasource (i.e., plugin) mapping file.

    Fields:
         - *attr_type:* Attribute Type; e.g., STATISTIC, RESULT_QUALITY, OBSERVED_PROPERTY; separate compound mappings with ':'
         - *basin3d_vocab:* The BASIN-3D vocabulary; separate compound mappings with ':'
         - *basin3d_desc:* The BASIN-3D vocabulary descriptions; objects or enum
         - *datasource_vocab:* The datasource vocabulary
         - *datasource_desc:* The datasource vocabulary description
         - *datasource:* The datasource of the mapping
    """
    attr_type: str
    basin3d_vocab: str
    basin3d_desc: list
    datasource_vocab: str
    datasource_desc: str
    datasource: DataSource = DataSource()

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.datasource_vocab


@dataclass
class MappedAttribute(JSONSerializable):
    """
    A data class for an attribute that is translated (i.e., mapped) from a datasource vocabulary to BASIN-3D vocabulary.
    Note that this model holds an AttributeMapping that maybe compound in nature; however this class specifies only one attribute types.
    For example, if the AttributeMapping is for a compound mapping of attribute types OBSERVED_PROPERTY:SAMPLING_MEDIUM,
    then the attr_type field would be either OBSERVED_PROPERTY or SAMPLING_MEDIUM but not both.

    Fields:
         - *attr_type:* Attribute Type; e.g., STATISTIC, RESULT_QUALITY, OBSERVED_PROPERTY, etc; single type only
         - *attr_mapping:* AttributeMapping as described in the datasource's (i.e., plugin's mapping file).
    """
    attr_type: MappedAttributeEnum
    attr_mapping: AttributeMapping

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.get_basin3d_vocab()

    def get_basin3d_vocab(self) -> Optional[str]:
        if self.attr_mapping.basin3d_vocab == NO_MAPPING_TEXT:
            return NO_MAPPING_TEXT
        attrs = self.attr_mapping.attr_type.split(MAPPING_DELIMITER)
        b3d_vocabs = self.attr_mapping.basin3d_vocab.split(MAPPING_DELIMITER)
        try:
            for attr, vocab in zip(attrs, b3d_vocabs):
                if attr == self.attr_type.upper():
                    return vocab
        except Exception as e:
            logger.error(f'Issue returning basin3d_vocab for MappedAttribute. THIS SHOULD NEVER HAPPEN. {e}')
        return None

    def get_basin3d_desc(self):
        b3d_descs = self.attr_mapping.basin3d_desc
        if not b3d_descs:
            return None
        attrs = self.attr_mapping.attr_type.split(MAPPING_DELIMITER)
        try:
            for attr, desc in zip(attrs, b3d_descs):
                if attr == self.attr_type.upper():
                    return desc
        except Exception as e:
            logger.error(f'Issue returning basin3d_desc for MappedAttribute. THIS SHOULD NEVER HAPPEN. {e}')
        return None

    def get_datasource_vocab(self) -> str:
        return self.attr_mapping.datasource_vocab

    def get_datasource_desc(self) -> str:
        return self.attr_mapping.datasource_desc


class Base(JSONSerializable):
    """
    Base synthesis model class. All classes that extend this are immutable.
    """

    def __init__(self, plugin_access, **kwargs):

        self._datasource_ids = None
        self._datasource = plugin_access and plugin_access.datasource
        self._id = None
        self._original_id = None

        if not isinstance(kwargs, dict):
            raise TypeError("Expected a dict")

        # any ids listed should have the DataSource.id_prefix
        if 'datasource_ids' in kwargs.keys():
            for id in kwargs['datasource_ids']:
                if kwargs[id]:
                    kwargs[id] = "{}-{}".format(self.datasource.id_prefix, kwargs[id])
        if "id" in kwargs and self._datasource:
            kwargs["original_id"] = kwargs["id"]
            kwargs["id"] = "{}-{}".format(self.datasource.id_prefix, kwargs["id"])

        # Now that we have massaged the incoming key/value pairs, let's
        # set them in the
        bad_attributes = []
        for key, value in kwargs.items():

            if not hasattr(self, key):
                bad_attributes.append(key)
            else:
                # If enum, get value
                setattr(self, key, isinstance(value, enum.Enum) and value.value or value)

        if len(bad_attributes) > 0:
            raise ValueError("Invalid argument(s) for {} : {}".format(self.__class__.__name__,
                                                                      ",".join(bad_attributes)))

        def __setattr__(self, *ignore_args):
            """
            This has been disabled.  The class is immutable

            :param ignore_args:
            :return:
            """
            raise AttributeError("{} is Immutable".format(self.__class__.__name__))

        def __delattr__(self, *ignore_args):
            """
               This has been disabled.  The class is immutable

               :param ignore_args:
               :return:
               """
            raise AttributeError("{} is Immutable".format(self.__class__.__name__))

        self.__setattr__ = __setattr__
        self.__delattr__ = __delattr__

    def _create_mapped_attributes(self, attr_type, attr_mappings):
        """

        :param attr_type:
        :param attr_mappings:
        :return:
        """

        def create_mapped_attribute(a_type, a_mapping):
            return MappedAttribute(a_type, a_mapping)

        if isinstance(attr_mappings, list) and attr_mappings:
            return list(map(create_mapped_attribute, repeat(attr_type), attr_mappings))

        elif attr_mappings:
            return create_mapped_attribute(attr_type, attr_mappings)

    def _translate_mapped_attributes(self, plugin_access, mapped_attrs, **kwargs):
        translated_attrs = translate_attributes(plugin_access, mapped_attrs, **kwargs)
        for attr in mapped_attrs:
            if attr in translated_attrs.keys() and translated_attrs[attr]:
                translated_attrs[attr] = self._create_mapped_attributes(attr.upper(), translated_attrs[attr])
        return translated_attrs

    @property
    def datasource_ids(self):
        return self._datasource_ids

    @datasource_ids.setter
    def datasource_ids(self, value):
        self._datasource_ids = value

    @property
    def datasource(self):
        return self._datasource

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def original_id(self):
        return self._original_id

    @original_id.setter
    def original_id(self, value):
        self._original_id = value


class Person(Base):
    """A person or organization"""

    def __init__(self, **kwargs):
        self._first_name: str = None
        self._last_name: str = None
        self._email: str = None
        self._institution: str = None
        self._role: str = None

        # Initialize after the attributes have been set
        super().__init__(None, **kwargs)

    @property
    def first_name(self) -> str:
        """First (given) name of person"""
        return self._first_name

    @first_name.setter
    def first_name(self, value: str):
        self._first_name = value

    @property
    def last_name(self) -> str:
        """Last (family) name"""
        return self._last_name

    @last_name.setter
    def last_name(self, value: str):
        self._last_name = value

    @property
    def email(self) -> str:
        """Email address"""
        return self._email

    @email.setter
    def email(self, value: str):
        self._email = value

    @property
    def institution(self) -> str:
        """Institution or organization name"""
        return self._institution

    @institution.setter
    def institution(self, value: str):
        self._institution = value

    @property
    def role(self) -> str:
        """Role of person in relation to responsibility"""
        return self._role

    @role.setter
    def role(self, value: str):
        self._role = value


class Coordinate(Base):
    """
    Top level coordinate class that holds :class:`AbsoluteCoordinate` or :class:`RepresentativeCoordinate`
    """

    def __init__(self, **kwargs):
        self._absolute: AbsoluteCoordinate = None
        self._representative: RepresentativeCoordinate = None

        # Initialize after the attributes have been set
        super().__init__(None, **kwargs)
        self.__validate__()

    def __validate__(self):
        """
        Validate the attributes
        """

        # enforce absolute is class AbsoluteCoordinate
        if self.absolute and not isinstance(self.absolute, AbsoluteCoordinate):
            raise TypeError("Coordinate.absolute attribute must be AbsoluteCoordinate object")

        # enforce representative class is RepresentativeCoordinate
        if self.representative and not isinstance(self.representative, RepresentativeCoordinate):
            raise TypeError("Coordinate.representative attribute must be RepresentativeCoordinate object")

        # enforce required coordinates: if only representative, then representative.representative_point is required
        # if self.absolute is None:
        #     if self.representative.representative_point is None:
        #         raise AttributeError("Representative_point is required if only representative coordinates are provided.")

    @property
    def absolute(self) -> 'AbsoluteCoordinate':
        """Absolute coordinate"""
        return self._absolute

    @absolute.setter
    def absolute(self, value: 'AbsoluteCoordinate'):
        self._absolute = value

    @property
    def representative(self) -> 'RepresentativeCoordinate':
        """Representative coordinate"""
        return self._representative

    @representative.setter
    def representative(self, value: 'RepresentativeCoordinate'):
        self._representative = value


class AbsoluteCoordinate(Base):
    """
    Absolute coordinate describes the geo-referenced location of a feature.
    Coordinates match the feature's shape. For example, a curve is a list of points.
    Currently collections of discrete points describing a feature are supported.

    """

    # Planned extension to better check point, curve, surface, solid shape-specific coordinates.
    # May want to include a type attribute akin to GeoJSON type
    # In future, reconsider the format of attributes to allow for more types of description (meshes, solids, etc)

    def __init__(self, **kwargs):
        self._horizontal_position: List[GeographicCoordinate] = []
        self._vertical_extent: List[AltitudeCoordinate] = []

        # Initialize after the attributes have been set
        super().__init__(None, **kwargs)
        self.__validate__()

    def __validate__(self):
        # require horizontal position and vertical extent to be lists
        if not isinstance(self.horizontal_position, (list, tuple, set)):  # check for better not iterable
            self.horizontal_position = [self.horizontal_position]

        if not isinstance(self.vertical_extent, (list, tuple, set)):
            self.vertical_extent = [self.vertical_extent]

        # ToDo: validate obj types
        for obj in self.horizontal_position:
            if not isinstance(obj, GeographicCoordinate):
                raise TypeError("Horizontal position must be instance of GeographicCoordinate")

        if len(self.vertical_extent) > 0:
            for obj in self.vertical_extent:
                if not isinstance(obj, AltitudeCoordinate):
                    raise TypeError("Vertical extent must be instance of AltitudeCoordinate")

            if len(self.horizontal_position) != len(self.vertical_extent):
                raise AttributeError("Lengths of horizontal positions and vertical extent must be equal.")

        # ToDo: add validation for shape coordinates.

    @property
    def horizontal_position(self) -> List['GeographicCoordinate']:
        """list of obj :class:`GeographicCoordinate`"""
        return self._horizontal_position

    @horizontal_position.setter
    def horizontal_position(self, value: List['GeographicCoordinate']):
        self._horizontal_position = value

    @property
    def vertical_extent(self) -> List['AltitudeCoordinate']:
        """list of obj :class:`AltitudeCoordinate`"""
        return self._vertical_extent

    @vertical_extent.setter
    def vertical_extent(self, value: List['AltitudeCoordinate']):
        self._vertical_extent = value


class RepresentativeCoordinate(Base):
    """
    Representative coordinates describe the location of a feature by a representative shape / location.
    For example, a study area may be represented by the center point.
    The veritical position from a reference position (e.g., height, depth) is also described in this class.
    Currently representative points are supported. The class is extendable to other forms of representing
    (e.g., diameter, area, side_length)
    Representative point types are also expandable as use cases require.
    """

    #: Placement of the representative point is the center of a local surface
    REPRESENTATIVE_POINT_TYPE_CENTER_LOCAL_SURFACE = "CENTER LOCAL SURFACE"

    #: Placement of the representative point is the upper left corner (northwest)
    REPRESENTATIVE_POINT_TYPE_UPPER_LEFT_CORNER = "UPPER LEFT CORNER"

    #: Placement of the representative point is the upper right corner (northeast)
    REPRESENTATIVE_POINT_TYPE_UPPER_RIGHT_CORNER = "UPPER RIGHT CORNER"

    #: Placement of the representative point is the lower left corner (southhwest)
    REPRESENTATIVE_POINT_TYPE_LOWER_LEFT_CORNER = "LOWER LEFT CORNER"

    #: Placement of the representative point is the lower right corner (northeast)
    REPRESENTATIVE_POINT_TYPE_LOWER_RIGHT_CORNER = "LOWER RIGHT CORNER"

    def __init__(self, **kwargs):
        self._representative_point: AbsoluteCoordinate = None
        self._representative_point_type: str = None
        self._vertical_position: DepthCoordinate = None

        # Initialize after the attributes have been set
        super().__init__(None, **kwargs)
        self.__validate__()

    def __validate__(self):
        """
        Validate attributes
        """

        # if representative point, require representative point type
        # if self.representative_point is not None:
        #     if self.representative_point_type is None:
        #         raise AttributeError("representative_point_type is required if representative_point provided.")

    @property
    def representative_point(self) -> AbsoluteCoordinate:
        """A point representation of the feature.
           obj :class:`AbsoluteCoordinate` for POINT"""
        return self._representative_point

    @representative_point.setter
    def representative_point(self, value: AbsoluteCoordinate):
        self._representative_point = value

    @property
    def representative_point_type(self) -> str:
        """The type of representative point relative to the feature's geometry
           Currently the point is assumed to be located at the local surface (CV).
           Use constants prefixed with `REPRESENTATIVE_POINT_TYPE_` """
        return self._representative_point_type

    @representative_point_type.setter
    def representative_point_type(self, value: str):
        self._representative_point_type = value

    @property
    def vertical_position(self) -> 'DepthCoordinate':
        """The vertical position of the feature from a reference position (e.g., height or depth).
           obj :class:`DepthCoordinate`"""
        return self._vertical_position

    @vertical_position.setter
    def vertical_position(self, value: 'DepthCoordinate'):
        self._vertical_position = value


class VerticalCoordinate(Base):
    """
    The vertical position of the feature (altitudes or depths).
    The reference frame or system is specified.

    """
    #: The distance above or below sea level (elevation)
    TYPE_ALTITUDE = "ALTITUDE"

    #: The distance above (height) or below (depth) of the local surface
    TYPE_DEPTH = "DEPTH"

    #: Distance in meters
    DISTANCE_UNITS_METERS = "meters"

    #: Distance in feet
    DISTANCE_UNITS_FEET = "feet"

    #: Explicit coordinate included with horizontal coordinates
    ENCODING_EXPLICIT = "EXPLICIT"

    #: Implicit coordinate
    ENCODING_IMPLICIT = "IMPLICIT"

    #: Attribute values
    ENCODING_ATTRIBUTE = "ATTRIBUTE"

    def __init__(self, **kwargs):
        self._value: float = None
        self._resolution: float = None
        self._distance_units: str = None
        self._encoding_method: str = None
        self._datum: str = None
        self._type: str = None

        # Initialize after the attributes have been set
        super().__init__(None, **kwargs)

    @property
    def type(self) -> str:
        """The type of veritical position :class:`VerticalCoordinate.TYPE_ALTITUDE` or
           :class:`VerticalCoordinate.TYPE_DEPTH`"""
        return self._type

    @type.setter
    def type(self, value: str):
        self._type = value

    @property
    def datum(self) -> str:
        """
        The reference coordinate system. Use constants prefixed with `DATUM_`
        """
        return self._datum

    @datum.setter
    def datum(self, value: str):
        self._datum = value

    @property
    def encoding_method(self) -> str:
        """The method for encoding the units of distance. Use constants prefixed with `ENCODING_` from :class:`VerticalCoordinate`"""
        return self._distance_units

    @encoding_method.setter
    def encoding_method(self, value: str):
        self._encoding_method = value

    @property
    def distance_units(self) -> str:
        """The unit of distance. It uses constants prefixed with `DISTANCE_UNITS_` from :class:`VerticalCoordinate`"""
        return self._distance_units

    @distance_units.setter
    def distance_units(self, value: str):
        self._distance_units = value

    @property
    def resolution(self) -> float:
        """The minimum distance possible between two adjacent
           depth values, expressed in Distance Units used for Depth"""
        return self._resolution

    @resolution.setter
    def resolution(self, value: float):
        self._resolution = value

    @property
    def value(self) -> float:
        """The vertical position value"""
        return self._value

    @value.setter
    def value(self, value: float):
        self._value = value


class AltitudeCoordinate(VerticalCoordinate):
    """
    An altitudinal vertical position (i.e., distance from sea level).
    The reference frame or system is specified. The term
    "altitude" is used instead of the common term "elevation" to conform to the terminology
    in Federal Information Processing Standards 70-1 and 173.
    """

    #: National Geodetic Vertical Datum of 1929
    DATUM_NGVD29 = "NGVD29"

    #: North American Vertical Datum of 1988
    DATUM_NAVD88 = "NAVD88"

    def __init__(self, **kwargs):
        self._datum: str = None

        # Initialize after the attributes have been set
        super().__init__(type=self.TYPE_ALTITUDE, **kwargs)

    @property
    def datum(self) -> str:
        """The reference coordinate system. Use constants prefixed with `DATUM_`"""
        return self._datum

    @datum.setter
    def datum(self, value: str):
        self._datum = value


class DepthCoordinate(VerticalCoordinate):
    """
    A depth vertical position (i.e., the height or depth from the specified reference position)
    The reference frame or system is specified.
    """

    #: Local surface
    DATUM_LOCAL_SURFACE = "LS"

    #: Mean sea level
    DATUM_MEAN_SEA_LEVEL = "MSL"

    def __init__(self, **kwargs):
        self._datum = None

        # Initialize after the attributes have been set
        super().__init__(type=self.TYPE_DEPTH, **kwargs)

    @property
    def datum(self) -> str:
        """The reference coordinate system. Use constants prefixed with `DATUM_`"""
        return self._datum

    @datum.setter
    def datum(self, value: str):
        self._datum = value


class HorizontalCoordinate(Base):
    """Generic XY coordinates for a point on earth (https://www.fgdc.gov/csdgmgraphical/spref.htm)"""

    #: World Geodetic System 1984 (WGS84)
    DATUM_WGS84 = "WGS84"

    #: North American Datum of 1983 (NAD 83)
    DATUM_NAD83 = "NAD83"

    #: North American Datum 1927 (NAD27)
    DATUM_NAD27 = "NAD27"

    #: The quantities of latitude and longitude which define the position of a
    #: point on the Earth's surface with respect to a reference spheroid.
    TYPE_GEOGRAPHIC = "GEOGRAPHIC"

    #: T plane-rectangular coordinate system usually based on, and
    #: mathematically adjusted to, a map projection so that geographic
    #: positions can be readily transformed to and from plane coordinates.
    TYPE_PLANAR_GRID = "PLANAR_GRID"

    #: Any right-handed planar coordinate system of which the z-axis
    #: coincides with a plumb line through the origin that locally is aligned with the surface of the Earth.
    TYPE_PLANAR_LOCAL = "PLANAR_LOCAL"

    #: The systematic representation of all or part of the surface of the Earth on a plane or developable surface.
    TYPE_PLANAR_MAP_PROJECTION = "PLANAR_MAP_PROJECTION"

    #: A description of any coordinate system that is not aligned with the surface of the Earth.
    TYPE_LOCAL = "LOCAL"

    def __init__(self, **kwargs):
        self._x: float = None
        self._y: float = None
        self._datum: str = None
        self._type: str = None

        # Initialize after the attributes have been set
        super().__init__(None, **kwargs)

    @property
    def x(self) -> float:
        """X Coordinate"""
        return self._x

    @x.setter
    def x(self, value: float):
        self._x = value

    @property
    def y(self) -> float:
        """Y Coordinate"""
        return self._y

    @y.setter
    def y(self, value: float):
        self._y = value

    @property
    def datum(self) -> str:
        """The reference coordinate system. Use constants prefixed with `DATUM_`"""
        return self._datum

    @datum.setter
    def datum(self, value: str):
        self._datum = value

    @property
    def type(self) -> str:
        """The type of horizontal coordinates. Use constants prefixed with `TYPE_` from :class:`HorizontalCoordinate`"""
        return self._type

    @type.setter
    def type(self, value: str):
        self._type = value


class GeographicCoordinate(HorizontalCoordinate):
    """
    The latitude and longitude which define the position of a point on
    the Earth's surface with respect to a reference spheroid.
    (https://www.fgdc.gov/csdgmgraphical/spref.htm)\
    """

    #: Decimal degrees
    UNITS_DEC_DEGREES = "DD"

    #: Decimal minutes
    UNITS_DEC_MINUTES = "DM"

    #: Decimal seconds
    UNITS_DEC_SECONDS = "DS"

    #: Degrees and decimal minutes
    UNITS_DEGREES_DEC_MINUTES = "DDM"

    #: Degrees, minutes, and decimal second
    UNITS_DEGREES_MIN_DEC_SECS = "DMDS"

    #: Radians
    UNITS_RADIANS = "Radians"

    #: Grads
    UNITS_GRADS = "Grads"

    UNITS = {UNITS_DEC_DEGREES: "Decimal degrees",
             UNITS_DEC_MINUTES: "Decimal minutes",
             UNITS_DEC_SECONDS: "Decimal seconds",
             UNITS_DEGREES_DEC_MINUTES: "Degrees and decimal minutes",
             UNITS_DEGREES_MIN_DEC_SECS: "Degrees, minutes, and decimal seconds",
             UNITS_RADIANS: UNITS_RADIANS,
             UNITS_GRADS: UNITS_GRADS
             }

    UNITS_DATA_TYPES = {UNITS_DEC_DEGREES: float,
                        UNITS_DEC_MINUTES: float,
                        UNITS_DEC_SECONDS: float,
                        UNITS_DEGREES_DEC_MINUTES: (int, float),
                        UNITS_DEGREES_MIN_DEC_SECS: (int, int, int),
                        UNITS_RADIANS: float,
                        UNITS_GRADS: float
                        }

    def __init__(self, **kwargs):
        self._units: str = None

        if "longitude" in kwargs:
            kwargs["x"] = kwargs["longitude"]
            kwargs.pop("longitude")
        if "latitude" in kwargs:
            kwargs["y"] = kwargs["latitude"]
            kwargs.pop("latitude")

        # Initialize after the attributes have been set
        super().__init__(type=self.TYPE_GEOGRAPHIC, **kwargs)
        self.__validate__()

    def __validate__(self):
        """
        Validate the attributes
        """

        # Validate that the units are valid
        if self.units not in self.UNITS.keys():
            units_list = ','.join(self.UNITS.keys())
            raise AttributeError(f'{self.units} is not a valid unit. Must be in {units_list}')

        # Validate that the unit values for x and y are the correct type
        units_data_type = self.UNITS_DATA_TYPES[self.units]
        for attribute in {'x', 'y'}:
            value = getattr(self, attribute)
            if self.units and value:
                if isinstance(units_data_type, tuple):

                    if not isinstance(value, tuple) or len(units_data_type) != len(value):
                        units_data_type_list = ','.join([x.__name__ for x in units_data_type])
                        raise TypeError(f'Value {value} for {self.__class__.__name__}.{attribute} '
                                        f'must be type {type(units_data_type).__name__}'
                                        f'({units_data_type_list}')
                    else:
                        for idx, v in enumerate(value):
                            if not isinstance(v, units_data_type[idx]):
                                units_data_type_list = ','.join([x.__name__ for x in units_data_type])
                                raise TypeError(f'Value {value} for {self.__class__.__name__}.{attribute} '
                                                f'must be type {type(units_data_type).__name__}'
                                                f'({units_data_type_list}')

                # FIXME units_data_type has a redline under it.
                elif not isinstance(value, units_data_type):
                    raise TypeError(f'Value {value} for {self.__class__.__name__}.{attribute} '
                                    f'must be type {units_data_type.__name__} not '
                                    f'({type(value).__name__}')

    @property
    def latitude(self) -> float:
        """Alias for Y Coordinate"""
        return self.y

    @property
    def longitude(self) -> float:
        """Alias for X Coordinate"""
        return self.x

    @property
    def units(self) -> str:
        """Latitude and longitude units. Use constants prefixed with `UNITS_`"""
        return self._units

    @units.setter
    def units(self, value: str):
        self._units = value


class Feature(Base):
    """
    A general feature upon which an observation can be made. Loosely after GF_Feature (ISO).
    """

    def __init__(self, plugin_access, **kwargs):
        self._id: str = None
        self._name: str = None
        self._description: str = None
        self._feature_type: str = None
        self._observed_properties: Union[List[MappedAttribute], List[str]] = None

        # Initialize after the attributes have been set
        super().__init__(plugin_access, **kwargs)

        if self.observed_properties:
            if not isinstance(self.observed_properties, list):
                logger.warning("observed_properties parameter not in expected list format")
            self.observed_properties = self._create_mapped_attributes('OBSERVED_PROPERTY', get_datasource_mapped_attribute(
                plugin_access, attr_type='OBSERVED_PROPERTY', datasource_vocab=self.observed_properties))
        else:
            self.observed_properties = None

    def __validate__(self):
        """
        Validate attributes
        """

        if self.feature_type is not None and self.feature_type not in FeatureTypeEnum.values():
            raise AttributeError("Feature attr feature_type must be FeatureTypeEnum.")

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self._id

    @property
    def id(self) -> str:
        """Unique identifier for the feature"""
        return self._id

    @id.setter
    def id(self, value: str):
        self._id = value

    @property
    def name(self) -> str:
        """A name for the feature"""
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def description(self) -> str:
        """The feature description"""
        return self._description

    @description.setter
    def description(self, value: str):
        self._description = value

    @property
    def feature_type(self) -> str:
        """The feature type. For a list of feature types see :class:`basin3d.schema.enum.FeatureTypeEnum`."""
        return self._feature_type

    @feature_type.setter
    def feature_type(self, value: str):
        self._feature_type = value

    @property
    def observed_properties(self) -> Union[List[MappedAttribute], List[str]]:
        """List of observed properties"""
        return self._observed_properties

    @observed_properties.setter
    def observed_properties(self, value: Union[List[MappedAttribute], List[str]]):
        self._observed_properties = value


class SamplingFeature(Feature):
    """
    A feature where sampling is conducted. OGC Observation & Measurements SF_SamplingFeature.
    """

    def __init__(self, plugin_access, **kwargs):
        self._related_sampling_feature_complex: List[SamplingFeature] = []

        # Initialize after the attributes have been set
        super().__init__(plugin_access, **kwargs)
        self.__validate__()

    # ToDo: validate items in lists
    def __validate__(self):
        if not isinstance(self.related_sampling_feature_complex, (list, tuple, set)):  # check for better not iterable
            self.related_sampling_feature_complex = [self.related_sampling_feature_complex]

    @property
    def related_sampling_feature_complex(self) -> List['SamplingFeature']:
        """List of related sampling features
           obj :class:`RelatedSamplingFeature`"""
        return self._related_sampling_feature_complex

    @related_sampling_feature_complex.setter
    def related_sampling_feature_complex(self, value: List['SamplingFeature']):
        self._related_sampling_feature_complex = value


class SpatialSamplingFeature(SamplingFeature):
    """
    A spatially-defined feature where sampling is conducted. OGC Observation & Measurements SF_SpatialSamplingFeature.
    """

    def __init__(self, plugin_access, **kwargs):
        self._shape: str = None
        self._coordinates: Coordinate = None

        # Initialize after the attributes have been set
        super().__init__(plugin_access, **kwargs)
        self.__validate__()

        # Set the shape dependent on feature_type
        for key, values in FEATURE_SHAPE_TYPES.items():
            if self.feature_type in values:
                self.shape = key

    def __validate__(self):
        """
        Require that feature_type is set
        """

        if self.feature_type is None:
            raise AttributeError("Feature attr feature_type must be indicated")

        if self.coordinates and not isinstance(self.coordinates, Coordinate):
            raise TypeError("coordinates must be Coordinate instance.")

        self._verify_coordinates_match_shape()

    def _verify_coordinates_match_shape(self):
        # Consider: invert logic in so that the coordinates specify the shape.
        error_msg = "Absolute coordinates do not match specified shape {}. ".format(self.shape)
        if self.coordinates and self.coordinates.absolute:
            if self.shape == SpatialSamplingShapes.SHAPE_POINT:
                if len(self.coordinates.absolute.horizontal_position) != 1:
                    raise AttributeError(error_msg + "Shape {} must have only one point."
                                         .format(SpatialSamplingShapes.SHAPE_POINT))
                else:
                    return
            if self.shape == SpatialSamplingShapes.SHAPE_SURFACE:
                if len(self.coordinates.absolute.horizontal_position) < 1 or \
                        self.coordinates.absolute.horizontal_position[0].x != \
                        self.coordinates.absolute.horizontal_position[-1].x or \
                        self.coordinates.absolute.horizontal_position[0].y != \
                        self.coordinates.absolute.horizontal_position[-1].y:
                    raise AttributeError(error_msg + "Shape {} must have more than one point. "
                                                     "The first and last points in the list must "
                                                     "be the same point.".format(SpatialSamplingShapes.SHAPE_SURFACE))
                else:
                    return
            if self.shape == SpatialSamplingShapes.SHAPE_CURVE:
                if len(self.coordinates.absolute.horizontal_position) < 1 or (
                        self.coordinates.absolute.horizontal_position[0].x ==
                        self.coordinates.absolute.horizontal_position[-1].x and
                        self.coordinates.absolute.horizontal_position[0].y ==
                        self.coordinates.absolute.horizontal_position[-1].y):
                    raise AttributeError(error_msg + "Shape {} must have more than one point. "
                                                     "The first and last points in the list must "
                                                     "NOT be the same point.".format(SpatialSamplingShapes.SHAPE_CURVE))
                else:
                    return
            # ToDo: distinguish solid from curve when altitude is included

    @property
    def coordinates(self) -> Coordinate:
        """Description of feature location. An instance of :class:`Coordinate`"""
        return self._coordinates

    @coordinates.setter
    def coordinates(self, value: Coordinate):
        self._coordinates = value

    @property
    def shape(self) -> str:
        """The shape of the feature. See :class:`basin3d.models.SpatialSamplingShapes`"""
        return self._shape

    @shape.setter
    def shape(self, value: str):
        self._shape = value


class MonitoringFeature(SpatialSamplingFeature):
    """
    A feature upon which monitoring is made. OGC Timeseries Profile OM_MonitoringFeature.
    """

    def __init__(self, plugin_access, **kwargs):
        self._description_reference: str = None
        self._related_party: List[Person] = []
        self._utc_offset: int = None

        # Initialize after the attributes have been set
        super().__init__(plugin_access, **kwargs)

    @property
    def description_reference(self) -> str:
        """Extra information about the Monitoring Feature"""
        return self._description_reference

    @description_reference.setter
    def description_reference(self, value):
        self._description_reference = value

    @property
    def related_party(self) -> List[Person]:
        """list of Person, people or organizations responsible for Feature.
           To be extended in future to full OGC Responsible_Party"""
        return self._related_party

    @related_party.setter
    def related_party(self, value: List[Person]):
        self._related_party = value

    @property
    def utc_offset(self) -> int:
        """Coordinated Universal Time (UTC) offset in hours (e.g. +/-9)"""
        return self._utc_offset

    @utc_offset.setter
    def utc_offset(self, value: int):
        self._utc_offset = value


class RelatedSamplingFeature(Base):
    """
    Class that represents a related sampling feature and its role relative to
    the sampling feature to which it is related. Spatial hierarchies of features
    are built by specifying related sampling features.

    Data model from OGC Observations and Measurements.
    """

    #: Sampling Feature is a parent
    ROLE_PARENT = "PARENT"

    ROLE_TYPES = [ROLE_PARENT]

    def __init__(self, plugin_access, **kwargs):
        self._related_sampling_feature: 'SamplingFeature' = None
        self._related_sampling_feature_type: str = None
        self._role: str = None

        # Initialize after the attributes have been set
        super().__init__(plugin_access, datasource_ids=['related_sampling_feature'], **kwargs)
        self.__validate__()

    def __validate__(self):
        """
        Validate attributes
        :return:
        """
        # ToDo: refactor this to not require type
        if self.related_sampling_feature_type is not None and \
                self.related_sampling_feature_type not in FeatureTypeEnum.values():
            raise AttributeError("RelatedSamplingFeature related_sampling_feature_type must be FeatureTypeEnum")

        if self.role is None:
            raise AttributeError("RelatedSamplingFeature role is required.")
        elif self.role not in RelatedSamplingFeature.ROLE_TYPES:
            raise AttributeError("RelatedSamplingFeature role must be one of predefined roles.")

    @property
    def related_sampling_feature(self) -> 'SamplingFeature':
        """A sampling feature relation"""
        return self._related_sampling_feature

    @related_sampling_feature.setter
    def related_sampling_feature(self, value: 'SamplingFeature'):
        self._related_sampling_feature = value

    @property
    def related_sampling_feature_type(self) -> str:
        """Feature type of the related sampling feature. See :class:`FeatureTypeEnum` for a list of types"""
        return self._related_sampling_feature_type

    @related_sampling_feature_type.setter
    def related_sampling_feature_type(self, value: str):
        self._related_sampling_feature_type = value

    @property
    def role(self) -> str:
        """Currently the only Related Sampling Feature role is a :class:`RelatedSamplingFeature.PARENT`"""
        return self._role

    @role.setter
    def role(self, value: str):
        self._role = value


class TimeValuePair(namedtuple('TimeValuePair', ['timestamp', 'value'])):
    """
    Tuple that represents a time value pair.  This will handle timestamp conversion

    `(timestamp, value)`
    """

    def __new__(cls, timestamp, value):
        # Handle epoch time
        if timestamp:
            timestamp_resolved = None
            if isinstance(timestamp, str) and timestamp.isdigit():
                timestamp_resolved = int(timestamp)
            elif isinstance(timestamp, Number):
                timestamp_resolved = timestamp

            if timestamp_resolved:
                timestamp = datetime.datetime.fromtimestamp(timestamp_resolved).isoformat()

        return super().__new__(cls, timestamp, value)


class Observation(Base):
    """

    OGC OM_Observation feature type. This is a parent class to which Mixins
        should be added to create observation types with metadata and result.
    """

    #: Measurement Time Value Pair Timeseries
    TYPE_MEASUREMENT_TVP_TIMESERIES = "MEASUREMENT_TVP_TIMESERIES"

    #: A measurement
    TYPE_MEASUREMENT = "MEASUREMENT"

    def __init__(self, plugin_access, **kwargs):
        self._id: str = None
        self._type: str = None
        self._utc_offset: int = None
        self._phenomenon_time: str = None
        self._observed_property: MappedAttribute = None
        self._feature_of_interest: MonitoringFeature = None
        self._feature_of_interest_type: FeatureTypeEnum = None
        self._result_quality: List[MappedAttribute] = []

        kwargs = self._translate_attributes(plugin_access, **kwargs)

        # Initialize after the attributes have been set
        super().__init__(plugin_access, **kwargs)
        self.__validate__()

    def __eq__(self, other):
        return self.id == other.id

    def __validate__(self):
        """
        Validate attributes
        """

        # Validate feature of interest type if present is class FeatureTypeEnum
        if self.feature_of_interest_type and self.feature_of_interest_type not in FeatureTypeEnum.values():
            raise AttributeError("feature_of_interest_type must be FeatureType")

    def _translate_attributes(self, plugin_access, **kwargs):
        # ToDo: see note with TimeseriesTVPObservation
        mapped_attrs = ('observed_property', 'result_quality')
        return self._translate_mapped_attributes(plugin_access, mapped_attrs, **kwargs)

    @property
    def id(self) -> str:
        """Unique observation identifier"""
        return self._id

    @id.setter
    def id(self, value: str):
        self._id = value

    @property
    def type(self) -> str:
        """Type of observation. Use constants prefixed with `TYPE_`"""
        return self._type

    @type.setter
    def type(self, value: str):
        self._type = value

    @property
    def utc_offset(self) -> int:
        """Coordinated Universal Time (UTC) offset in hours (e.g. +/-9)"""
        return self._utc_offset

    @utc_offset.setter
    def utc_offset(self, value: int):
        self._utc_offset = value

    @property
    def phenomenon_time(self) -> str:
        """datetime of the observation (required OGC attribute timePhenomenon).
           For timeseries, start and end datetimes can be provided."""
        return self._phenomenon_time

    @phenomenon_time.setter
    def phenomenon_time(self, value: str):
        self._phenomenon_time = value

    @property
    def observed_property(self) -> 'MappedAttribute':
        """The property that was observed"""
        return self._observed_property

    @observed_property.setter
    def observed_property(self, value: 'MappedAttribute'):
        self._observed_property = value

    @property
    def feature_of_interest(self) -> 'MonitoringFeature':
        """The feature on which the observed property was observed"""
        return self._feature_of_interest

    @feature_of_interest.setter
    def feature_of_interest(self, value: 'MonitoringFeature'):
        self._feature_of_interest = value

    @property
    def feature_of_interest_type(self) -> 'FeatureTypeEnum':
        """The type of feature that was observed. See :class:`basin3d.models.FeatureTypeEnum`"""
        return self._feature_of_interest_type

    @feature_of_interest_type.setter
    def feature_of_interest_type(self, value: 'FeatureTypeEnum'):
        self._feature_of_interest_type = value

    @property
    def result_quality(self) -> List['MappedAttribute']:
        """The result quality assessment. See :class:`ResultQuality`"""
        return self._result_quality

    @result_quality.setter
    def result_quality(self, value: List['MappedAttribute']):
        self._result_quality = value


class TimeMetadataMixin(object):
    """
    Metadata attributes for Observations with a time
    """
    #: Observation taken at the start
    TIME_REFERENCE_START = "START"

    #: Observation taken in the middle
    TIME_REFERENCE_MIDDLE = "MIDDLE"

    #: Observation taken at the end
    TIME_REFERENCE_END = "END"

    def __init__(self, *args, **kwargs):
        self._aggregation_duration: MappedAttribute = None
        self._time_reference_position: str = None

        # Instantiate the serializer superclass
        super(TimeMetadataMixin, self).__init__(*args, **kwargs)

    @property
    def aggregation_duration(self) -> 'MappedAttribute':
        """Time period represented by the observation. Follows OGC TM_PeriodDuration.
           Use constants prefixed with `AGGREGATION_DURATION` from :class:`TimeseriesMetadataMixin`"""
        return self._aggregation_duration

    @aggregation_duration.setter
    def aggregation_duration(self, value: 'MappedAttribute'):
        self._aggregation_duration = value

    @property
    def time_reference_position(self) -> str:
        """Position of timestamp in aggregated_duration. Encompassed as part of OGC interpolationType.
           Use constants prefixed with `TIME_REFERENCE` from :class:`TimeseriesMetadataMixin`"""
        return self._time_reference_position

    @time_reference_position.setter
    def time_reference_position(self, value: str):
        self._time_reference_position = value


class MeasurementMetadataMixin(object):
    """
    Metadata attributes for Observations type Measurement
    """

    def __init__(self, *args, **kwargs):
        self._sampling_medium: MappedAttribute = None
        self._statistic: MappedAttribute = None

        # Instantiate the serializer superclass
        super(MeasurementMetadataMixin, self).__init__(*args, **kwargs)

    @property
    def sampling_medium(self) -> 'MappedAttribute':
        """Sampling medium in which the observed property was measured"""
        return self._sampling_medium

    @sampling_medium.setter
    def sampling_medium(self, value: 'MappedAttribute'):
        self._sampling_medium = value

    @property
    def statistic(self) -> 'MappedAttribute':
        """The statistical property of the observation result. Use constants prefixed with `STATISTIC_` from :class:`MeasurementMetadataMixin`"""
        return self._statistic

    @statistic.setter
    def statistic(self, value: 'MappedAttribute'):
        self._statistic = value


class ResultListTVP(Base):
    """
    Result Point Float
    """
    def __init__(self, plugin_access, **kwargs):
        self._value: List['TimeValuePair'] = []
        self._result_quality: List['MappedAttribute'] = []

        # translate quality
        kwargs = self._translate_attributes(plugin_access, **kwargs)

        # Initialize after the attributes have been set
        super().__init__(plugin_access, **kwargs)

    def _translate_attributes(self, plugin_access, **kwargs):
        # ToDo: see note with TimeseriesTVPObservation
        mapped_attrs = ('result_quality', )
        return self._translate_mapped_attributes(plugin_access, mapped_attrs, **kwargs)

    @property
    def value(self) -> List['TimeValuePair']:
        """Result that was measured"""
        return self._value

    @value.setter
    def value(self, value: List['TimeValuePair']):
        self._value = value

    @property
    def result_quality(self) -> List['MappedAttribute']:
        """Result that was measured"""
        return self._result_quality

    @result_quality.setter
    def result_quality(self, value: List['MappedAttribute']):
        self._result_quality = value


class MeasurementTimeseriesTVPResultMixin(object):
    """
    Result Mixin: Measurement Timeseries TimeValuePair
    """

    def __init__(self, *args, **kwargs):
        self._result: 'ResultListTVP' = None
        self._unit_of_measurement: str = None

        # Instantiate the serializer superclass
        super(MeasurementTimeseriesTVPResultMixin, self).__init__(*args, **kwargs)

    @property
    def result(self) -> 'ResultListTVP':
        """A list of results """
        return self._result

    @result.setter
    def result(self, value: 'ResultListTVP'):
        self._result = value

    @property
    def unit_of_measurement(self) -> str:
        """Unit of measurement"""
        return self._unit_of_measurement

    @unit_of_measurement.setter
    def unit_of_measurement(self, value: str):
        self._unit_of_measurement = value


class ResultPointFloat(Base):
    """
    Result Point Float
    """
    def __init__(self, plugin_access, **kwargs):
        self._value: float = None
        self._result_quality: 'MappedAttribute' = None

        # translate quality
        kwargs = self._translate_attributes(plugin_access, **kwargs)

        # Initialize after the attributes have been set
        super().__init__(None, **kwargs)

    def _translate_attributes(self, plugin_access, **kwargs):
        # ToDo: see note with TimeseriesTVPObservation
        mapped_attrs = ('result_quality', )
        return self._translate_mapped_attributes(plugin_access, mapped_attrs, **kwargs)

    @property
    def value(self) -> float:
        """Result that was measured"""
        return self._value

    @value.setter
    def value(self, value: float):
        self._value = value

    @property
    def result_quality(self) -> 'MappedAttribute':
        """Result that was measured"""
        return self._result_quality

    @result_quality.setter
    def result_quality(self, value: 'MappedAttribute'):
        self._result_quality = value


class MeasurementResultMixin(object):
    """
    Result Mixin: Measurement
    """

    def __init__(self, **kwargs):
        # self._result_value: float = None
        self._result: 'ResultPointFloat' = None
        self._unit_of_measurement: str = None

        # Instantiate the serializer superclass
        super().__init__(**kwargs)

    @property
    def result(self) -> 'ResultPointFloat':
        """Result"""
        return self._result

    @result.setter
    def result(self, value: 'ResultPointFloat'):
        self._result = value

    @property
    def unit_of_measurement(self) -> str:
        """Unit of measurement"""
        return self._unit_of_measurement

    @unit_of_measurement.setter
    def unit_of_measurement(self, value: str):
        self._unit_of_measurement = value


class MeasurementTimeseriesTVPObservation(TimeMetadataMixin, MeasurementMetadataMixin,
                                          MeasurementTimeseriesTVPResultMixin, Observation):
    """
    Series of measurement (numerical) observations in TVP format grouped by time (i.e., a timeseries).
    Anything specified at the group level automatically applies to the individual observation.
    """

    # NOTE: Position Observation (the one inheriting from Base) last in the inheritance list.
    def __init__(self, plugin_access, **kwargs):
        kwargs["type"] = self.TYPE_MEASUREMENT_TVP_TIMESERIES

        self._translate_attributes(plugin_access, **kwargs)

        # Initialize after the attributes have been set
        super(MeasurementTimeseriesTVPObservation, self).__init__(plugin_access, **kwargs)

    def __eq__(self, other):
        return self.id == other.id

    def _translate_attributes(self, plugin_access, **kwargs):
        # ToDo: Introspect which attributes are MappedAttributes. Cannot do this easily b/c using @property decorator.
        # For a simple dataclass, typing.get_type_hints would work. As properties, the fget for each property can be introspected with typing.get_type_hints
        mapped_attrs = ('observed_property', 'statistic', 'aggregation_duration', 'result_quality', 'sampling_medium')
        return self._translate_mapped_attributes(plugin_access, mapped_attrs, **kwargs)




"""

.. currentmodule:: basin3d.plugins.epa

:platform: Unix, Mac
:synopsis: EPA Water Quality eXchange Plugin Definition
:module author: Danielle Christianson <dschristianson@lbl.gov>


* :class:`EPADataSourcePlugin` - This Data Source plugin maps the EPA Water Quality Exchange Data Source to the BASIN-3D Models
"""
import csv
import json
import os
import requests
import urllib.parse

from copy import deepcopy
from typing import Iterator, List, Optional, Union
from datetime import datetime as dt, date

from basin3d.core import monitor
from basin3d.core.schema.enum import FeatureTypeEnum, MappedAttributeEnum
from basin3d.core.schema.query import QueryMeasurementTimeseriesTVP, QueryMonitoringFeature
from basin3d.core.access import get_url
from basin3d.core.models import AbsoluteCoordinate, RepresentativeCoordinate, Coordinate, GeographicCoordinate, \
    DepthCoordinate, VerticalCoordinate, MeasurementTimeseriesTVPObservation, MonitoringFeature, \
    TimeMetadataMixin, TimeValuePair, ResultListTVP, AttributeMapping
from basin3d.core.plugin import DataSourcePluginAccess, DataSourcePluginPoint, basin3d_plugin
from basin3d.core.translate import get_datasource_mapped_attribute
from basin3d.core.types import SpatialSamplingShapes

logger = monitor.get_logger(__name__)


# https://www.epa.gov/waterdata/storage-and-retrieval-and-water-quality-exchange-domain-services-and-downloads
# Download link: https://cdx.epa.gov/wqx/download/DomainValues/TimeZone.csv
# Accessed 2023-10-03
TIMEZONE_MAP = {
    "ADT": {"utc_offset": "-03:00", "name": "Atlantic Daylight Time"},  # -3,8/11/2006 10:57:50 AM
    "AHST": {"utc_offset": "-10:00", "name": "Alaska-Hawaii Standard Time (*retired: >1983 use AKST)"},  # -10,8/11/2006 10:57:50 AM
    "AKDT": {"utc_offset": "-08:00", "name": "Alaska Daylight Time"},  # -8,8/11/2006 10:57:50 AM
    "AKST": {"utc_offset": "-09:00", "name": "Alaska Standard Time"},  # -9,8/11/2006 10:57:50 AM
    "AST": {"utc_offset": "-04:00", "name": "Atlantic Standard Time"},  # -4,8/11/2006 10:57:50 AM
    "BST": {"utc_offset": "-11:00", "name": "Bering Standard Time (*retired: >1983 use HAST)"},  # -11,8/11/2006 10:57:50 AM
    "CDT": {"utc_offset": "-05:00", "name": "Central Daylight Time"},  # -5,8/11/2006 10:57:50 AM
    "CEST": {"utc_offset": "+02:00", "name": "Sweden Daylight Time or Central European Standard Time"},  # 2,7/12/2016 9:39:33 AM
    "CET": {"utc_offset": "+01:00", "name": "Stockholm Sweden Time or Central European Time"},  # 1,7/12/2016 9:39:33 AM
    "CST": {"utc_offset": "-06:00", "name": "Central Standard Time"},  # -6,8/11/2006 10:57:50 AM
    "EDT": {"utc_offset": "-04:00", "name": "Eastern Daylight Time"},  # -4,8/11/2006 10:57:50 AM
    "EST": {"utc_offset": "-05:00", "name": "Eastern Standard Time"},  # -5,8/11/2006 10:57:50 AM
    "GMT": {"utc_offset": "+00:00", "name": "Greenwich Mean Time"},  # 0,6/19/2009 11:19:08 AM
    "GST": {"utc_offset": "+10:00", "name": "Guam Standard Time Zone (also Chamorro Standard Time)"},  # 10,8/11/2006 10:57:50 AM
    "HADT": {"utc_offset": "-09:00", "name": "Hawaii-Aleutian Daylight Time"},  # -9,8/11/2006 10:57:50 AM
    "HAST": {"utc_offset": "-10:00", "name": "Hawaii-Aleutian Standard Time"},  # -10,8/11/2006 10:57:50 AM
    "KST": {"utc_offset": "+09:00", "name": "Korea Standard Time"},  # 9,7/12/2016 9:39:33 AM
    "MDT": {"utc_offset": "-06:00", "name": "Mountain Daylight Time"},  # -6,8/11/2006 10:57:50 AM
    "MST": {"utc_offset": "-07:00", "name": "Mountain Standard Time"},  # -7,8/11/2006 10:57:50 AM
    "NDT": {"utc_offset": "-02:00", "name": "Newfoundland Daylight Time"},  # -2.5,8/11/2006 10:57:50 AM
    "NST": {"utc_offset": "-03:00", "name": "Newfoundland Standard Time"},  # -3.5,8/11/2006 10:57:50 AM
    "PDT": {"utc_offset": "-07:00", "name": "Pacific Daylight Time"},  # -7,8/11/2006 10:57:50 AM
    "PST": {"utc_offset": "-08:00", "name": "Pacific Standard Time"},  # -8,8/11/2006 10:57:50 AM
    "SST": {"utc_offset": "-11:00", "name": "American Samoa Standard Time"},  # -11,8/11/2006 10:57:50 AM
    "UTC": {"utc_offset": "+00:00", "name": "Coordinated Universal Time"},  # 0,1/15/2010 1:10:00 PM
    "YST": {"utc_offset": "-09:00", "name": "Yukon Standard Time: U.S.-Yukatat (*retired: >1983 use AKST)"}  # -9,8/11/2006 10:57:50 AM
}

EPA_WQP_API_VERSION = os.environ.get('EPA_WQP_API_VERSION', '2.2')

# mapping of EPA WQP API v2.2 to v3.0
# https://www.epa.gov/waterdata/water-quality-portal-quick-reference-guide
# accessed 2024-09-23
FIELD_NAMES = {
    'loc_id': {'2.2': 'MonitoringLocationIdentifier', '3.0': 'Location_Identifier'},
    'lat': {'2.2': 'LatitudeMeasure', '3.0': 'Location_Latitude'},
    'long': {'2.2': 'LongitudeMeasure', '3.0': 'Location_Longitude'},
    'org_name': {'2.2': 'OrganizationFormalName', '3.0': 'Org_FormalName'},
    'loc_name': {'2.2': 'MonitoringLocationName', '3.0': 'Location_Name'},
    'huc_8_code': {'2.2': 'HUCEightDigitCode', '3.0': 'Location_HUCEightDigitCode'},
    'provider': {'2.2': 'ProviderName', '3.0': 'ProviderName'},
    'phys_chem_results': {'2.2': 'resultPhysChem', '3.0': 'fullPhysChem'},
    'observed_property': {'2.2': 'CharacteristicName', '3.0': 'Result_Characteristic'},
    'sampling_media': {'2.2': 'ActivityMediaName', '3.0': 'Activity_Media'},
    'sample_fraction': {'2.2': 'ResultSampleFractionText', '3.0': 'Result_SampleFraction'},
    'depth_height': {'2.2': 'ResultDepthHeightMeasure/MeasureValue', '3.0': 'ResultDepthHeight_Measure'},
    'aggregation_duration': {'2.2': 'ResultTimeBasisText', '3.0': 'Result_TimeBasis'},
    'sampling_temp': {'2.2': 'ResultTemperatureBasisText', '3.0': 'Result_MeasureTemperatureBasis'},
    'statistic': {'2.2': 'StatisticalBaseCode', '3.0': 'Result_StatisticalBase'},
    'start_date': {'2.2': 'ActivityStartDate', '3.0': 'Activity_StartDate'},
    'start_time': {'2.2': 'ActivityStartTime/Time', '3.0': 'Activity_StartTime'},
    'start_time_zone': {'2.2': 'ActivityStartTime/TimeZoneCode', '3.0': 'Activity_StartTimeZone'},
    'end_date': {'2.2': 'ActivityEndDate', '3.0': 'Activity_EndDate'},
    'end_time': {'2.2': 'ActivityEndTime/Time', '3.0': 'Activity_EndTime'},
    'end_time_zone': {'2.2': 'ActivityEndTime/TimeZoneCode', '3.0': 'Activity_EndTimeZone'},
    'result': {'2.2': 'ResultMeasureValue', '3.0': 'Result_Measure'},
    'result_unit': {'2.2': 'ResultMeasure/MeasureUnitCode', '3.0': 'Result_MeasureUnit'},
    'result_status': {'2.2': 'ResultStatusIdentifier', '3.0': 'Result_MeasureStatusIdentifier'},
    'result_type': {'2.2': 'ResultValueTypeName', '3.0': 'Result_MeasureType'},
    'activity_type': {'2.2': 'ActivityTypeCode', '3.0': 'Activity_TypeCode'},
    'depth_height_unit': {'2.2': 'ResultDepthHeightMeasure/MeasureUnitCode', '3.0': 'ResultDepthHeight_MeasureUnit'},
    'depth_height_ref': {'2.2': 'ResultDepthAltitudeReferencePointText', '3.0': 'ResultDepthHeight_AltitudeReferencePoint'}
}

EPA_GEOSERVER_WFS_TIMEOUT_LIMIT = os.environ.get('EPA_GEOSERVER_WFS_TIMEOUT_LIMIT', 5)


def _post_wqp_search(search_type: str, query_data: dict,
                     api_version: str = EPA_WQP_API_VERSION) -> Optional[requests.Response]:
    """
    Request the locations information
    :param search_type: EPA WQP Data Profile type: Physical / Chemical (See FIELD_NAMES key 'phys_chem_results' for vocab, or Station (for locations)
    :param query_data: query in WQP vocabulary for the request
    :param api_version: EPA WQP API version: '2.2' and '3.0' currently supported; otherwise the url will be invalid.
    """
    url_api_version = 'INVALID_API_VERSION'

    if api_version == '2.2':
        url_api_version = 'data'
    elif api_version == '3.0':
        url_api_version = 'wqx3'

    url_with_params = f'https://www.waterqualitydata.us/{url_api_version}/{search_type}/search?mimeType=csv'
    headers = {'content-type': 'application/json'}
    data = {'providers': ['STORET']}
    data.update(query_data)
    logger.info(f'url: {url_with_params} and payload {str(data)}')

    try:
        data_json = json.dumps(data)
        resp = requests.post(url_with_params, data_json, headers=headers, verify=True, stream=True)
        logger.info(f'response: {resp.status_code}')
        return resp
    except Exception as e:
        logger.error(f'Unsuccessful attempt to query WQP; {e}')
        return None


def _get_location_info(loc_type: str, loc_list: list, synthesis_messages: list = []) -> list:
    """
    Get location information. First try the GeoServer WFS endpoint and if is not available
       within the timeout limit, fail over to a WQP Station request.
    :param loc_type: siteid or huc
    :param loc_list: list of location identifiers
    :param synthesis_messages: list of messages to report warning / error messages
    """
    loc_info_list = []
    fail_over = False
    msg: Optional[str] = None

    # try the ogc approach
    try:
        if loc_type == 'siteid':
            id_list_str = _make_mf_query_str(loc_list)
            loc_info_list = _get_location_info_ogc(id_list_str, synthesis_messages)
        else:
            for huc in loc_list:
                huc_query_str = f'huc%3A{huc}'
                loc_info_results = _get_location_info_ogc(huc_query_str, synthesis_messages)
                loc_info_list.extend(loc_info_results)
    # if fails, stream the csv file
    except TimeoutError as e:
        fail_over = True
        msg = f'WFS Geoserver timed out, fail over to WQP Station request\nError: {e}'
    except Exception as e:
        fail_over = True
        msg = f'WFS Geoserver did not respond appropriately, fail over to WQP Station request\nError: {e}'

    if fail_over:
        logger.warning(msg)
        synthesis_messages.append(msg)
        print(msg)
        loc_info_list.extend(_get_location_info_csv(loc_type, loc_list, synthesis_messages))

    return loc_info_list


def _get_location_info_csv(loc_type: str, loc_list: list, synthesis_messages: list) -> list:
    """
    Get location information from the EPA Data Source using the Station WQP request type
    :param loc_type: siteid or huc
    :param loc_list: list of location identifiers
    :param synthesis_messages: list of messages to report warning / error messages
    """

    api_version = EPA_WQP_API_VERSION
    loc_info: List[dict] = []
    wqp_response = _post_wqp_search('Station', {loc_type: loc_list}, api_version)

    if not wqp_response or (wqp_response.status_code and wqp_response.status_code != 200):
        msg = 'No or invalid response to WQP Station request'
        logger.error(msg)
        synthesis_messages.append(msg)
        return loc_info

    wqp_csv = _get_csv_dict_reader(wqp_response)

    # Build a dictionary object that looks like the OGC WFS structure so have one parsing code
    for idx, row_dict in enumerate(wqp_csv):
        properties_name = row_dict.get(FIELD_NAMES['loc_id'][api_version])

        if not properties_name:
            msg = f'Row {idx} of the response csv does not have an identifier, skipping'
            logger.warning(msg)
            synthesis_messages.append(msg)
            continue

        geometry_coord_lat = row_dict.get(FIELD_NAMES['lat'][api_version])
        geometry_coord_long = row_dict.get(FIELD_NAMES['long'][api_version])

        if not geometry_coord_lat or not geometry_coord_long:
            msg = f'Row {idx} with identifier {properties_name} does not have both lat / long coordinates. skipping'
            logger.warning(msg)
            synthesis_messages.append(msg)
            continue

        properties_org_name = row_dict.get(FIELD_NAMES['org_name'][api_version])
        properties_loc_name = row_dict.get(FIELD_NAMES['loc_name'][api_version])
        properties_huc8 = row_dict.get(FIELD_NAMES['huc_8_code'][api_version])
        properties_provider = row_dict.get(FIELD_NAMES['provider'][api_version])

        loc_info.append({
            'geometry': {'type': 'Point', 'coordinates': [geometry_coord_lat, geometry_coord_long]},
            'properties': {'name': properties_name, 'locName': properties_loc_name,
                           'huc8': properties_huc8, 'orgName': properties_org_name,
                           'provider': properties_provider}})

    return loc_info


def _get_location_info_ogc(loc_str: str, synthesis_messages: list) -> list:
    """
    Get location information from the EPA Data Source for specified locations from WFS in OGC format
    :param loc_str: query string for the GeoServer WFS request
    :param synthesis_messages: list of messages to report warning / error messages
    """
    loc_info_list = []
    url = ('https://www.waterqualitydata.us/ogcservices/wfs/?request=GetFeature&service=wfs&version=2.0.0&typeNames=wqp_sites'
           f'&SEARCHPARAMS={loc_str}%3Bproviders%3ASTORET&outputFormat=application%2Fjson')

    # ToDo: Enhancements -- stream and chunk for expected large returns, chunk site_ids into multiple calls if large list
    result = get_url(url, timeout=EPA_GEOSERVER_WFS_TIMEOUT_LIMIT)

    if result and result.status_code and result.status_code == 200:
        try:
            result_data = json.loads(result.content)
            loc_info_list = result_data.get('features', [])
        except Exception as e:
            msg = f'EPA WQX {url} result could not be parsed {e}'
            logger.warning(msg)
            synthesis_messages.append(msg)
    else:
        error_code = 'NO RESPONSE'
        if result and result.status_code:
            error_code = result.status_code
        msg = f'EPA WQX {url} returned error code {error_code}'
        logger.warning(msg)
        synthesis_messages.append(msg)

    return loc_info_list


def _make_monitoring_feature_object(plugin_access: DataSourcePluginAccess, loc_info_obj: dict) -> Optional[MonitoringFeature]:
    """
    Generate a monitoring feature object from the EPA location info
    :param loc_info_obj:
    :return:
    """
    loc_geometry_type = None
    loc_geometry = loc_info_obj.get('geometry')
    if loc_geometry:
        loc_geometry_type = loc_geometry.get('type')
    if not loc_geometry_type or loc_geometry_type.upper() != FeatureTypeEnum.POINT.value:
        return None

    loc_properties = loc_info_obj.get('properties')
    if not loc_properties:
        return None

    loc_id = loc_properties.get('name')
    loc_name = loc_properties.get('locName')
    huc8 = loc_properties.get('huc8')
    org = loc_properties.get('orgName')
    provider = loc_properties.get('provider')

    desc = f'Location is part of USGS huc {huc8}; organization {org}; provider {plugin_access.datasource.id} {provider}'

    monitoring_feature = MonitoringFeature(
        plugin_access,
        id=loc_id,
        name=loc_name,
        description=desc,
        feature_type=FeatureTypeEnum.POINT,
        shape=SpatialSamplingShapes.SHAPE_POINT)

    coord: list = loc_geometry.get('coordinates')
    if coord:
        monitoring_feature.coordinates = Coordinate(
            absolute=AbsoluteCoordinate(horizontal_position=GeographicCoordinate(
                **{"latitude": float(coord[1]),
                   "longitude": float(coord[0]),
                   "units": GeographicCoordinate.UNITS_DEC_DEGREES})))

    return monitoring_feature


def _make_mf_query_str(query_mf: list) -> str:
    """
    Create a monitoring feature query string for the EPA locations REST API
    :param query_mf:
    :return:
    """
    id_list = []
    for idx, mf_id in enumerate(query_mf):
        url_encoded_id = urllib.parse.quote(mf_id)
        if idx == 0:
            id_list.append(f'siteid%3A{url_encoded_id}')
            continue
        id_list.append(f'%7C{url_encoded_id}')
    return ''.join(id_list)


def _reformat_date_for_epa_query(b3d_query_date: date) -> str:
    """
    Change BASIN-3D date object to EPA format: MM-DD-YYYY
    :param b3d_query_date:
    :return:
    """
    date_str = str(b3d_query_date)  # YYYY-MM-DD
    date_str_pieces = date_str.split('-')
    return f'{date_str_pieces[1]}-{date_str_pieces[2]}-{date_str_pieces[0]}'


class EPAMonitoringFeatureAccess(DataSourcePluginAccess):
    """
    Access for mapping EPA Water Quality Exchange sampling locations to
    :class:`~basin3d.core.models.MonitoringFeature` objects.

    Using WFS OGC Mapping service: '~/ogcservices/wfs/?request=GetFeature&service=wfs&version=2.0.0&typeNames=wqp_sites&SEARCHPARAMS=providers%3ASTORET&outputFormat=application%2Fjson'
    Backup: WQP Station endpoint: 'https://www.waterqualitydata.us/{url_api_version}/Station/search?mimeType=csv'

    ============== === ====================================================
    EPA WQX            BASIN-3D
    ============== === ====================================================
    siteid          >> :class:`basin3d.core.models.MonitoringFeature`
    ============== === ====================================================

    """
    synthesis_model_class = MonitoringFeature

    def list(self, query: QueryMonitoringFeature) -> Iterator[Optional[MonitoringFeature]]:
        """
        Return measurement locations based on the query parameters
        Either a parent_feature list or monitoring_feature list should be specified.
        USGS hucs can be specified for parent_feature parameters (e.g., EPA-14020001)

        :param query: MonitoringFeature query
        :return:
        """
        synthesis_messages: List[str] = []

        feature_type = isinstance(query.feature_type, FeatureTypeEnum) and query.feature_type.value or query.feature_type

        if feature_type and feature_type not in EPADataSourcePlugin.feature_types:
            msg = f'Feature type {feature_type} not supported by {self.datasource.name}.'
            synthesis_messages.append(msg)
            logger.warning(msg)
        elif not query.parent_feature and not query.monitoring_feature:
            msg = f'{self.datasource.name} requires either a parent feature or monitoring feature be specified in the query.'
            synthesis_messages.append(msg)
            logger.warning(msg)
        elif query.parent_feature and query.monitoring_feature:
            msg = f'{self.datasource.name} does not support querying monitoring features by both parent_feature (huc) and monitoring_feature (list of ids).'
            synthesis_messages.append(msg)
            logger.warning(msg)
        else:
            loc_list = []
            loc_type = None
            if query.parent_feature:
                for parent_feature in query.parent_feature:
                    if len(parent_feature) not in [2, 4, 6, 8, 10, 12] or not parent_feature.isdigit():
                        msg = f'{self.datasource.name}: {parent_feature} does not appear to be a valid USGS huc: 2, 4, 6, 8, 10, or 12-digit code.'
                        synthesis_messages.append(msg)
                        logger.warning(msg)
                        continue
                    wildcard = '*'
                    if len(parent_feature) >= 8:
                        wildcard = ''
                    loc_list.append(f'{parent_feature}{wildcard}')
                if loc_list:
                    loc_type = 'huc'

            elif query.monitoring_feature:
                loc_list = deepcopy(query.monitoring_feature)
                loc_type = 'siteid'

            if loc_type:
                loc_info_list = _get_location_info(loc_type, loc_list, synthesis_messages)  # type: ignore
                for loc_info in loc_info_list:
                    mf_obj = _make_monitoring_feature_object(self, loc_info)
                    if mf_obj:
                        yield mf_obj

        return StopIteration(synthesis_messages)  # type: ignore

    def get(self, query: QueryMonitoringFeature):
        """
        Get a EPA measurement location based on the id
        :param query: MonitoringFeature query, id must be specified
        :return:
        """
        # query.id will always be a string at this point with validation upstream, thus ignoring the type checking

        loc_info = _get_location_info('siteid', [query.id])

        if loc_info:
            if len(loc_info) > 1:
                msg = f'{self.datasource.id}: monitoring feature query by id {query.id} yielded too many site_id results from WQP. Cannot report single site.'
                logger.warning(msg)
                return None
            return _make_monitoring_feature_object(self, loc_info[0])

        return None


def _get_csv_dict_reader(epa_response: requests.Response) -> csv.DictReader:
    """
    Generate a DictReader from the EPA data response. Isolate the functionality so that it can easily be mocked for testing.
    :param epa_response:
    :return:
    """
    return csv.DictReader(epa_response.iter_lines(decode_unicode=True))


def _convert_distance_unit_to_meters(unit: str) -> Optional[float]:
    """
    Distance unit converter for depth / height values.
    :param unit: the measurement unit variable
    :return:
    """
    multiplier = None
    if unit == 'ft':
        multiplier = 0.3048
    elif unit == 'in':
        multiplier = 0.0254
    elif unit == 'cm':
        multiplier = 0.01
    elif unit == 'm':
        multiplier = 1
    return multiplier


def _parse_epa_results_phys_chem(wqp_response: requests.Response, query: QueryMeasurementTimeseriesTVP, op_map: dict,
                                 results: dict, synthesis_messages: list, api_version: str = EPA_WQP_API_VERSION) -> set:
    """
    Helper to parse the resultsPhysChem data.
    See https://www.waterqualitydata.us/portal_userguide/#table-7-sample-results-physicalchemical-result-retrieval-metadata

    Combo to find repeated measures in time (api_version = '2.2' | '3.0')
      MonitoringLocationIdentifier | Location_Identifier -- endpoint query
      CharacteristicName | Result_Characteristic -- endpoint query
      ActivityMediaName | Activity_Media -- endpoint query
      ResultSampleFractionText | Result_SampleFraction
      ResultDepthHeightMeasure/MeasureValue | ResultDepthHeight_Measure
      ResultTimeBasisText | Result_TimeBasis -- filter response | aggregation_duration, assume no value is instantaneous
      ResultTemperatureBasisText | Result_MeasureTemperatureBasis
      StatisticalBaseCode | Result_StatisticalBase -- filter response | statistic

    Time (api_version = '2.2' | '3.0')
      NOTE: combo these start time values to report in the TVP time. Set time position to start.
      ActivityStartDate | Activity_StartDate-- endpoint query
      ActivityStartTime/Time | Activity_StartTime
      ActivityStartTime/TimeZoneCode | Activity_StartTimeZone

      # Using start date only, but including here for completeness (api_version = '2.2' | '3.0')
      ActivityEndDate | Activity_EndDate -- endpoint query
      ActivityEndTime/Time | Activity_EndTime
      ActivityEndTime/TimeZoneCode | Activity_EndTimeZone

    Measurement results (api_version = '2.2' | '3.0')
      ResultMeasureValue | Result_Measure = value
      ResultMeasure/MeasureUnitCode | Result_MeasureUnit = unit
      ResultStatusIdentifier | Result_MeasureStatusIdentifier -- filter response | result_quality
      ResultValueTypeName | Result_MeasureType -- filter response for Estimated ONLY | result_quality

    Additional info (api_version = '2.2' | '3.0')
      ActivityTypeCode | Activity_TypeCode, e.g. Field Msr/Obs, Sample-Routine
      ResultDepthHeightMeasure/MeasureUnitCode | ResultDepthHeight_MeasureUnit
      ResultDepthAltitudeReferencePointText | ResultDepthHeight_AltitudeReferencePoint

    :param wqp_response:
    :param query:
    :param op_map: observed property map of datasource variable vocab to basin vocab
    :param results:
    :param synthesis_messages:
    :return:
    """

    # ToDo: get estimated value from attribute mappings directly.
    EPA_ESTIMATED_VOCAB = 'Estimated'

    def translate_empty_value(value):
        """
        Turn empty string values into None
        :param value:
        :return:
        """
        if value == '':
            return None
        return value

    def make_timestamp(row_dict: dict) -> Optional[str]:
        start_date = row_dict.get(FIELD_NAMES['start_date'][api_version])
        if not start_date:
            return None

        start_timestamp_str = f'{start_date}'
        timestamp_str_format = '%Y-%m-%d'

        start_time = row_dict.get(FIELD_NAMES['start_time'][api_version])
        if start_time:
            timestamp_str_format += ' %H:%M:%S'
            start_timestamp_str += f' {start_time}'

        start_time_zone = row_dict.get(FIELD_NAMES['start_time_zone'][api_version])
        if start_time_zone:
            mapped_time_zone = TIMEZONE_MAP.get(start_time_zone)
            if mapped_time_zone:
                start_time_zone_int = mapped_time_zone.get("utc_offset")
                if start_time_zone_int:
                    start_timestamp_str += start_time_zone_int
                    timestamp_str_format += '%z'

        try:
            return dt.strptime(start_timestamp_str, timestamp_str_format).isoformat(sep='T')
        except Exception as e:
            # ToDo: enhancement, include more info in the msg
            msg = f'Could not parse and convert start timestamp: {start_timestamp_str}; {e}'
            synthesis_messages.append(msg)
            logger.warning(msg)
            return None

    def str_to_numeric(value) -> Optional[Union[float, int]]:
        if not translate_empty_value(value):
            return None

        is_float = True
        if '.' not in value:
            is_float = False

        try:
            if is_float:
                return float(value)
            return int(value)
        except Exception:
            # ToDo: enhancement, include more info in the msg
            msg = f'Could not parse expected numerical measurement value {value}'
            synthesis_messages.append(msg)
            logger.warning(msg)
            return None

    def is_in_query(row_dict: dict, query: QueryMeasurementTimeseriesTVP) -> bool:
        """
        (api version = '2.2' | '3.0')
        ResultTimeBasisText | Result_TimeBasis -- filter response | aggregation_duration, assume empty value is instantaneous
        StatisticalBaseCode | Result_StatisticalBase -- filter response | statistic
        ResultStatusIdentifier | Result_MeasureStatusIdentifier -- filter response | result_quality
        ResultValueTypeName | Result_MeasureType -- filter response for Estimated ONLY | result_quality

        :param row_dict:
        :param query:
        :return:
        """

        if query.aggregation_duration and query.aggregation_duration[0] != 'NONE':
            ds_aggregation_duration = row_dict.get(FIELD_NAMES['aggregation_duration'][api_version])
            if not ds_aggregation_duration or ds_aggregation_duration and ds_aggregation_duration not in query.aggregation_duration:
                return False
        elif query.aggregation_duration[0] == 'NONE':
            if row_dict.get(FIELD_NAMES['aggregation_duration'][api_version]):
                return False

        if query.statistic:
            ds_statistic = row_dict.get(FIELD_NAMES['statistic'][api_version])
            if not ds_statistic or ds_statistic and ds_statistic not in query.statistic:
                return False

        # EPA estimated vocab is in vocabulary list separate from the rest of the result quality terms.
        # Revisit if BASIN-3D adds a ResultValueType
        if query.result_quality and EPA_ESTIMATED_VOCAB in query.result_quality:
            ds_result_value_type = row_dict.get(FIELD_NAMES['result_type'][api_version])
            if ds_result_value_type == EPA_ESTIMATED_VOCAB:
                return True

        if query.result_quality:
            ds_result_value_type = row_dict.get(FIELD_NAMES['result_type'][api_version])
            if ds_result_value_type == EPA_ESTIMATED_VOCAB:
                return False
            ds_result_quality = row_dict.get(FIELD_NAMES['result_status'][api_version])
            if not ds_result_quality or ds_result_quality and ds_result_quality not in query.result_quality:
                return False

        return True

    # --------------- End of internal helper functions

    mf_set = set()

    # wqp_csv = csv.DictReader(wqp_response.iter_lines(decode_unicode=True))
    wqp_csv = _get_csv_dict_reader(wqp_response)
    for row_dict in wqp_csv:
        # filter out any results that don't match the query fields that were not specified in the call to WQP
        if not is_in_query(row_dict, query):
            continue

        # make the dictionary key
        mf_id = translate_empty_value(row_dict.get(FIELD_NAMES['loc_id'][api_version]))
        # epa_observed_property = translate_empty_value(row_dict.get('CharacteristicName'))
        epa_observed_property = translate_empty_value(row_dict.get(FIELD_NAMES['observed_property'][api_version]))
        measurement = str_to_numeric(row_dict.get(FIELD_NAMES['result'][api_version]))

        # check first set of required info
        if any([x is None for x in [mf_id, epa_observed_property, measurement]]):
            continue

        # get the basin3d vocab for synthesis and skip if no match
        basin3d_vocab = op_map.get(epa_observed_property)
        if not basin3d_vocab:
            continue

        mf_set.add(mf_id)

        # time
        start_timestamp = make_timestamp(row_dict)

        # if the time fields cannot be parsed, move on.
        if not start_timestamp:
            continue

        tvp = TimeValuePair(timestamp=start_timestamp, value=measurement)

        epa_sampling_medium = row_dict.get(FIELD_NAMES['sampling_media'][api_version])
        epa_sample_fraction = row_dict.get(FIELD_NAMES['sample_fraction'][api_version])
        # Note there are also Activity DepthHeight fields
        epa_depth_height = row_dict.get(FIELD_NAMES['depth_height'][api_version])
        epa_depth_height_unit = row_dict.get(FIELD_NAMES['depth_height_unit'][api_version])
        epa_depth_height_ref = row_dict.get(FIELD_NAMES['depth_height_ref'][api_version])
        epa_aggregation_duration = row_dict.get(FIELD_NAMES['aggregation_duration'][api_version])
        epa_sampling_temp = row_dict.get(FIELD_NAMES['sampling_temp'][api_version])
        epa_statistic = row_dict.get(FIELD_NAMES['statistic'][api_version])
        epa_unit = translate_empty_value(row_dict.get(FIELD_NAMES['result_unit'][api_version]))

        # result_quality lives in 2 fields so first look in one field that has priority over the others.
        # If not, translate the the other field.
        epa_result_quality = translate_empty_value(row_dict.get(FIELD_NAMES['result_type'][api_version]))
        if epa_result_quality is None or epa_result_quality != EPA_ESTIMATED_VOCAB:
            epa_result_quality = translate_empty_value(row_dict.get(FIELD_NAMES['result_status'][api_version]))

        # The data return are a mix of siteids, variables, dates, etc because the data are not timeseries.
        # so make a unique key out of the fields that would define a single time series. Add timestamps / data to that
        # unique timeseries as the return is streamed.
        result_key = (f'{mf_id}-{basin3d_vocab}-{epa_sampling_medium}-{epa_unit}-{epa_depth_height}'
                      f'-{epa_depth_height_unit}-{epa_aggregation_duration}-{epa_statistic}-{epa_sampling_temp}')
        combo_dict = results.setdefault(result_key, {})
        if 'metadata' not in combo_dict.keys():
            id_info = f'{mf_id}-{epa_observed_property}'
            if translate_empty_value(epa_sample_fraction):
                id_info = f'{id_info}-sample_fraction:{epa_sample_fraction}'
            if translate_empty_value(epa_sampling_temp):
                id_info = f'{id_info}-activity_temp:{epa_sampling_temp}'
            if translate_empty_value(epa_depth_height_ref):
                id_info = f'{id_info}-depth_ref:{epa_depth_height_ref}'

            combo_dict['metadata'] = {
                'id': id_info,
                'mf_id': mf_id,
                'sampling_medium': translate_empty_value(epa_sampling_medium),
                'aggregation_duration': translate_empty_value(epa_aggregation_duration),
                'result_quality': set(),
                'statistic': translate_empty_value(epa_statistic),
                'unit_of_measurement': epa_unit,
                'height_depth': str_to_numeric(epa_depth_height),
                'height_depth_unit': translate_empty_value(epa_depth_height_unit),
                'observed_property': epa_observed_property
            }

        if epa_result_quality:
            combo_dict['metadata']['result_quality'].add(epa_result_quality)

        combo_dict.setdefault('result_tvp', []).append(tvp)
        combo_dict.setdefault('result_quality', []).append(epa_result_quality)

    return mf_set


class EPAMeasurementTimeseriesTVPObservationAccess(DataSourcePluginAccess):
    """
    Water Quality Portal Service: https://www.waterqualitydata.us/

    Access for mapping EPA water quality data to
    :class:`~basin3d.core.models.MeasurementTimeseriesTVPObservation` objects.

    """

    synthesis_model_class = MeasurementTimeseriesTVPObservation

    def list(self, query: QueryMeasurementTimeseriesTVP):
        """
        Generate EPA Measurement Timeseries TVP object. The data are not really time series. But we are treating them as such.

        :param query:
        :return:
        """
        api_version = EPA_WQP_API_VERSION

        synthesis_messages: List[str] = []

        params = {'dataProfile': FIELD_NAMES['phys_chem_results'][api_version],
                  'siteid': query.monitoring_feature,
                  'characteristicName': query.observed_property,
                  'startDateLo': _reformat_date_for_epa_query(query.start_date)}

        if query.sampling_medium:
            params.update({"sampleMedia": query.sampling_medium})

        if query.end_date:
            params.update({"startDateHi": _reformat_date_for_epa_query(query.end_date)})

        # get data from resultPhysChem; possible extension biological results (sample based data)
        epa_data = _post_wqp_search('Result', params, api_version)

        mf_list = []
        if epa_data and epa_data.status_code == 200:
            # Because the data are all mixed together and we support multiple mappings,
            #   group the variables by the BASIN-3D vocab. Create a look up store for the variables in the query.
            op_map = self._get_observed_property_map(query.observed_property)

            results = {}  # type: ignore[var-annotated]
            mf_set = _parse_epa_results_phys_chem(epa_data, query, op_map, results, synthesis_messages, api_version)
            mf_list = list(mf_set)

        if mf_list:
            # mf_query_str = _make_mf_query_str(mf_list)
            loc_info = _get_location_info('siteid', mf_list, synthesis_messages)
            loc_info_store = {}
            for loc_info_obj in loc_info:
                loc_properties = loc_info_obj.get('properties')
                if not loc_properties:
                    continue
                loc_id = loc_properties.get('name')
                if loc_id:
                    loc_info_store[loc_id] = loc_info_obj

            # loop thru the results
            for result in results.values():
                metadata = result.get('metadata')

                mf_id = metadata.get('mf_id')
                monitoring_feature_info = loc_info_store.get(mf_id)
                if not monitoring_feature_info:
                    msg = f'{self.datasource.id}: Location information for {mf_id} could not be found.'
                    logger.debug(msg)
                    continue

                mf_obj = _make_monitoring_feature_object(self, monitoring_feature_info)

                if not mf_obj:
                    continue

                if metadata.get('height_depth'):
                    depth_height_unit = metadata.get('height_depth_unit')
                    unit_multiplier = _convert_distance_unit_to_meters(depth_height_unit)
                    if depth_height_unit and unit_multiplier:
                        depth_height = metadata.get('height_depth')
                        depth_height *= unit_multiplier

                        if mf_obj.coordinates:
                            mf_obj.coordinates.representative = RepresentativeCoordinate(
                                vertical_position=DepthCoordinate(
                                    value=depth_height,
                                    distance_units=VerticalCoordinate.DISTANCE_UNITS_METERS))

                metadata_result_quality = None
                if metadata.get('result_quality'):
                    metadata_result_quality = list(metadata.get('result_quality'))

                # deal with result values
                measurement_result_quality = result.get('result_quality')
                if any(measurement_result_quality):
                    tvp_result = ResultListTVP(self, value=result.get('result_tvp'), result_quality=measurement_result_quality)
                else:
                    tvp_result = ResultListTVP(self, value=result.get('result_tvp'))

                meas_tvp_obj = MeasurementTimeseriesTVPObservation(
                    plugin_access=self,
                    id=metadata.get('id'),
                    observed_property=metadata.get('observed_property'),
                    feature_of_interest=mf_obj,
                    feature_of_interest_type=mf_obj.shape,
                    time_reference_position=TimeMetadataMixin.TIME_REFERENCE_START,
                    result_quality=metadata_result_quality,
                    result=tvp_result,
                    statistic=metadata.get('statistic'),
                    aggregation_duration=metadata.get('aggregation_duration'),
                    sampling_medium=metadata.get('sampling_medium'),
                    unit_of_measurement=metadata.get('unit_of_measurement')
                )

                yield meas_tvp_obj

        else:
            msg = f'{self.datasource.id}: No resultPhysChem results matched the query: {str(params)}'
            synthesis_messages.append(msg)

        return StopIteration(synthesis_messages)

    def _get_observed_property_map(self, observed_properties) -> dict:
        """
        Get a map of data source vocabulary to basin 3d vocabulary to group observations.
        :param observed_properties: list, observed properties
        :return:
        """
        op_map = {}
        for observed_property in observed_properties:
            attr_mapping: AttributeMapping = get_datasource_mapped_attribute(self, MappedAttributeEnum.OBSERVED_PROPERTY.value, observed_property)
            basin3d_vocab = attr_mapping.basin3d_vocab
            op_map[observed_property] = basin3d_vocab

        return op_map


@basin3d_plugin
class EPADataSourcePlugin(DataSourcePluginPoint):
    title = 'EPA Water Quality eXchange Data Source Plugin'
    plugin_access_classes = (EPAMonitoringFeatureAccess, EPAMeasurementTimeseriesTVPObservationAccess)

    # IF this list grows, EPAMonitoringFeature logic may need to be updated.
    feature_types = ['POINT']

    class DataSourceMeta:
        """
        This is an internal metadata class for defining additional :class:`~basin3d.models.DataSource`
        attributes.

        **Attributes:**
            - *id* - unique id short name
            - *name* - human friendly name (more descriptive)
            - *location* - resource location
            - *id_prefix* - id prefix to make model object ids unique across plugins
            - *credentials_format* - if the data source requires authentication, this is where the
                format of the stored credentials is defined.

        """
        # Citation info: National Water Quality Monitoring Council, YYYY, Water Quality Portal, accessed mm, dd, yyyy, hyperlink_for_query, https://doi.org/10.5066/P9QRKUVJ.

        # Data Source attributes
        id = 'EPA'  # unique id for the datasource
        location = 'https://www.waterqualitydata.us'
        id_prefix = 'EPA'
        name = 'EPA Water Quality eXchange'  # Human Friendly Data Source Name


"""

.. currentmodule:: basin3d.core.sqlalchemy_models

:platform: Unix, Mac
:synopsis: BASIN-3D ``DataSource`` catalog classes
:module author: Valerie C. Hendrix <vchendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
import csv
from sqlalchemy.exc import SQLAlchemyError, NoResultFound, IntegrityError

import basin3d
from basin3d.core import monitor, models
from importlib import resources
from inspect import getmodule
from string import whitespace
from typing import Iterator, List, Optional, Union

from basin3d.core.schema.enum import MAPPING_DELIMITER, NO_MAPPING_TEXT, MappedAttributeEnum, \
    set_mapped_attribute_enum_type, BaseEnum
from basin3d.core import sqlalchemy_models

logger = monitor.get_logger(__name__)
"""The logger for the catalog"""

DATASOURCE_NONE = models.DataSource(name="", location="", id_prefix="", id="", credentials={})
"""The default DataSource object when a DataSource is not found"""


class CatalogException(Exception):
    """The exception class for the Catalog"""
    pass


class CatalogBase:

    def __init__(self, variable_filename: str = 'basin3d_observed_property_vocabulary.csv'):
        """
        Initialize the catalog

        :param variable_filename: the filename of the observed property vocabulary
        """
        self.variable_dir = 'basin3d.data'  # location of the variable file
        self.variable_filename = variable_filename  # observed property vocabulary filename
        self._plugin_ids: List[str] = []  # list of plugin ids (updated on initialization)

    def initialize(self, plugin_list: list):
        """
        Initialize the catalog. This method should be called before any other method.


        :param plugin_list: list of plugins
        :return: tinyDB object
        """
        if not self.is_initialized():
            logger.debug(f"Initializing {self.__class__.__name__} metadata catalog ")

            def _get_plugin_id(plugin):
                """Get the plugin id from the plugin"""
                try:
                    if isinstance(plugin, str):
                        return basin3d.core.plugin.PluginMount.plugins[plugin].get_meta().id
                    return plugin.get_meta().id
                except Exception:
                    logger.error(
                        'Could not retrieve plugin_id. Check that plugin is configured / registered properly.')
                    raise CatalogException

            self._plugin_ids = plugin_list and [_get_plugin_id(plugin) for plugin in plugin_list if plugin] or []

            # initiate db
            self._init_catalog(plugin_ids=self._plugin_ids)

            # generate variable store
            self._gen_variable_store()

            # Load plugins, get the mapping file, and insert into the catalog datastore
            for plugin in plugin_list:

                try:
                    # if the plugin doesn't exist then getting the id should fail
                    plugin_id = plugin.get_meta().id
                    logger.info(f"Loading metadata catalog for Plugin {plugin_id}")
                except Exception:
                    logger.error(
                        'Could not retrieve plugin_id. Check that plugin is configured / registered properly.')
                    raise CatalogException

                mapping_filename = f'{plugin_id.lower()}_mapping.csv'
                plugin_module = getmodule(plugin)
                if not plugin_module:
                    logger.error(f'Plugin {plugin_id} cannot be found')
                    raise CatalogException(f'Plugin {plugin_id} cannot be found')

                plugin_package = ".".join(plugin_module.__name__.split(".")[0:-1])
                # checking for the mapping file might occur elsewhere or look different depending on where plugins live
                has_mapping_file = resources.is_resource(plugin_package, mapping_filename)

                if not has_mapping_file:
                    logger.error(f'Plugin {plugin_id} does not have variable mapping file {mapping_filename}')
                    raise CatalogException(f'Plugin {plugin_id} does not have variable mapping file {mapping_filename}')

                datasource = plugin.get_datasource()
                self._process_plugin_attr_mapping(plugin, mapping_filename, datasource)

            logger.info(f"Initialized {self.__class__.__name__} metadata catalog ")

    # ---------------------------------
    # Methods to be over-witten by children classes

    def _init_catalog(self, **kwargs):
        """
        Initialize the catalog database

        """
        raise NotImplementedError

    def _insert(self, record):
        """
        Insert the record

        :param record: the record to insert
        """
        raise NotImplementedError

    def _get_datasource(self, datasource_id) -> Optional[basin3d.core.models.DataSource]:
        """
        Access a single datasource

        :param datasource_id: the datasource_identifier
        """
        raise NotImplementedError

    def _get_observed_property(self, basin3d_vocab) -> Optional[basin3d.core.models.ObservedProperty]:
        """
        Access a single observed property variable

        :param basin3d_vocab: the observed property variable identifier
        """
        raise NotImplementedError

    def _get_attribute_mapping(self, datasource_id, attr_type, basin3d_vocab, datasource_vocab, **kwargs) -> Optional[
            basin3d.core.models.AttributeMapping]:
        """
        Access a single attribute mapping

        :param datasource_id: the datasource identifier, e.g., plugin_id
        :param attr_type: the attribute type, e.g., STATISTIC
        :param basin3d_vocab: the BASIN-3D vocabulary, e.g., Max
        :param datasource_vocab: the datasource vocabulary, e.g., max
        :return: the attribute mapping
        """
        raise NotImplementedError

    def is_initialized(self) -> bool:
        """Has the catalog been initialized?"""
        raise NotImplementedError

    def find_observed_property(self, basin3d_vocab) -> Optional[basin3d.core.models.ObservedProperty]:
        """
        Return the :class:`basin3d.models.ObservedProperty` object for the BASIN-3D vocabulary specified.

        :param basin3d_vocab: BASIN-3D vocabulary
        :return: a :class:`basin3d.models.ObservedProperty` object
        """
        raise NotImplementedError

    def find_observed_properties(self, basin3d_vocab=None) -> Iterator[Optional[basin3d.core.models.ObservedProperty]]:
        """
        Report the observed_properties available based on the BASIN-3D vocabularies specified. If no BASIN-3D vocabularies are specified, then return all observed properties available.

        :param basin3d_vocab: list of the BASIN-3D observed properties
        :return: generator that yields :class:`basin3d.models.ObservedProperty` objects
        """
        raise NotImplementedError

    def find_datasource_attribute_mapping(self, datasource_id, attr_type, attr_vocab) -> Optional[
            basin3d.core.models.AttributeMapping]:
        """
        Find the datasource attribute vocabulary to BASIN-3D mapping given a specific datasource_id, attr_type, and datasource attr_vocab.

        :param: datasource_id: the datasource identifier
        :param: attr_type: attribute type
        :param: datasource_vocab: the datasource attribute vocabulary
        :return: a :class:`basin3d.models.AttributeMapping` object
        """
        raise NotImplementedError

    def find_attribute_mappings(self, datasource, attr_type, attr_vocab, from_basin3d) -> Iterator[
            basin3d.core.models.AttributeMapping]:
        """
        Find the list of attribute mappings given the specified fields.
        Exact matches are returned (see attr_vocab formats below for BASIN-3D vocab nuances).
        If no fields are specified, all registered attribute mappings will be returned.

        :param datasource: the datasource object
        :param attr_type: the attribute type
        :param attr_vocab: the attribute vocabulary, the formats are one of the following:
                           1) datasource vocab that is a complete vocab
                           2) BASIN-3D vocab that is a complete vocab for a given attr_type regardless if it is compound or not.
                           3) BASIN-3D vocab that is compound and fully specified for each attr_type either with the complete vocab or with wildcards.
        :param from_basin3d: boolean that says whether the attr_vocab is a BASIN-3D vocabulary. If not, then this a datasource vocabulary.
        :return: generator that yields :class:`basin3d.models.AttributeMapping` objects
        """
        # class extensions should use _validate_fine_attribute_mappings_arguments
        raise NotImplementedError

    # --------------------------------------

    def _gen_variable_store(self):
        """
        Generate a variable store. Loads this from the provided vocabulary CSV file.

        """
        fields = ['basin3d_vocab', 'description', 'categories', 'units']
        # location of mapping file might change.
        with resources.open_text(self.variable_dir, self.variable_filename) as variable_file:

            reader = csv.DictReader(variable_file)

            # For now, force a specific file format; could change later to just require specific field names
            if reader.fieldnames != fields:
                logger.critical(f'{self.variable_filename} is not in correct format. Cannot create catalog.')
                raise CatalogException

            for row in reader:
                basin3d_vocab = row[fields[0]]

                if self._get_observed_property(basin3d_vocab) is not None:
                    logger.warning(f'Duplicate BASIN-3D variable {basin3d_vocab} found. Skipping duplicate.')
                    continue

                categories_list = []
                if row[fields[2]]:
                    categories_str = row[fields[2]]
                    categories_list = categories_str.split(',')
                    categories_list = [category.strip(whitespace) for category in categories_list]

                observed_property_variable = basin3d.core.models.ObservedProperty(
                    basin3d_vocab=basin3d_vocab,
                    full_name=row[fields[1]],
                    categories=categories_list,
                    units=row[fields[3]])

                self._insert(observed_property_variable)

    def _get_attribute_enum(self, str_value, enum_type):
        """
        Get the enum for a given string value

        :param str_value: string value to convert to enum, e.g., Max
        :param enum_type: the enum type, e.g., StatisticEnum
        :return:
        """
        enum_value = None
        try:
            enum_value = getattr(enum_type, str_value)
        except AttributeError:
            pass
        return enum_value

    def _process_plugin_attr_mapping(self, plugin, filename: str, datasource: basin3d.core.models.DataSource):
        """
        Process the plugin attribute mapping file

        :param plugin: the plugin, e.g., the plugin object
        :param filename: the filename of the mapping file, e.g., 'plugin_mapping.csv'
        :param datasource: the datasource object
        :return:
        """
        fields = ['attr_type', 'basin3d_vocab', 'datasource_vocab', 'datasource_desc']

        plugin_id = plugin.get_meta().id
        plugin_module = getmodule(plugin)
        if not plugin_module:
            logger.error(f'Plugin {plugin_id} cannot be found')
            raise CatalogException(f'Plugin {plugin_id} cannot be found')

        plugin_package = ".".join(plugin_module.__name__.split(".")[0:-1])

        with resources.open_text(plugin_package, filename) as attr_file:
            logger.debug(f"Mapping file {filename} for plugin package {plugin_package}")
            reader = csv.DictReader(attr_file)

            # For now, force a specific file format; could change later to just require specific field names
            if reader.fieldnames != fields:
                logger.critical(
                    f'Plugin {plugin_id}: {filename} is not in correct format. Cannot create catalog.')
                raise CatalogException(
                    f'Plugin {plugin_id}: {filename} is not in correct format. Cannot create catalog.')

            # loop thru the file
            for row in reader:
                attr_type = row[fields[0]]
                basin3d_vocab = row[fields[1]]
                datasource_vocab = row[fields[2]]
                datasource_desc = row[fields[3]]

                # ToDo: add some validation if Cat does not have it in her reader
                if self._get_attribute_mapping(datasource.id, attr_type, basin3d_vocab, datasource_vocab) is not None:
                    logger.warning(
                        f'Plugin {plugin_id}: Duplicate BASIN-3D attr detected. Cannot handle duplicate mappings yet. '
                        f'Skipping datasource attribute {datasource_vocab}.')

                # build the basin3d_description from objects (e.g., for OBSERVED_PROPERTY) or enums
                basin3d_desc = []
                for a_type, b3d_vocab in zip(attr_type.split(MAPPING_DELIMITER),
                                             basin3d_vocab.split(MAPPING_DELIMITER)):
                    valid_type_vocab = True

                    b3d_attr_type = None
                    try:
                        b3d_attr_type = getattr(MappedAttributeEnum, a_type)
                        if b3d_attr_type == MappedAttributeEnum.OBSERVED_PROPERTY:
                            b3d_desc = self._get_observed_property(b3d_vocab)
                        else:
                            b3d_enum_type = set_mapped_attribute_enum_type(b3d_attr_type)
                            b3d_desc = self._get_attribute_enum(b3d_vocab, b3d_enum_type)
                    except AttributeError:
                        logger.warning(
                            f'Datasource {datasource.id}: Attribute type {a_type} is not supported. Skipping mapping.')
                        valid_type_vocab = False
                        break

                    if not b3d_desc:
                        logger.warning(
                            f'Datasource {datasource.id}: basin3d_vocab {b3d_vocab} for attr_type {a_type} '
                            f'is not a valid BASIN-3D vocabulary. Skipping attribute mapping.')
                        valid_type_vocab = False
                        break

                    basin3d_desc.append(b3d_desc)

                if not valid_type_vocab:
                    continue

                attr_mapping = basin3d.core.models.AttributeMapping(
                    attr_type=attr_type,
                    basin3d_vocab=basin3d_vocab,
                    basin3d_desc=basin3d_desc,
                    datasource_vocab=datasource_vocab,
                    datasource_desc=datasource_desc,
                    datasource=datasource)

                self._insert(attr_mapping)
                logger.debug(f"{datasource.id}: Mapped {attr_type} {datasource_vocab} to {basin3d_vocab}")


class CatalogSqlAlchemy(CatalogBase):

    def __init__(self, variable_filename: str = 'basin3d_observed_property_vocabulary.csv'):
        super().__init__(variable_filename)

        # Clear the database, on initialization
        basin3d.core.sqlalchemy_models.clear_database()

    def is_initialized(self) -> bool:
        """Has the catalog been initialized?"""

        session = sqlalchemy_models.Session()

        try:
            datasources = session.query(sqlalchemy_models.DataSource).count()
            if isinstance(datasources, int):
                return datasources > 0
            logger.debug('Catalog not initialized')
            return False
        except SQLAlchemyError as e:
            logger.error(f'SQLAlchemy error: {e}')
            return False
        finally:
            session.close()

    def _convert_observed_property(self, sqlalchemy_opv) -> Optional[basin3d.core.models.ObservedProperty]:
        """
        Convert SQL Alchemy observed property variable to basin3d

        :param sqlalchemy_opv:
        :return: basin3d observed property
        """

        return basin3d.core.models.ObservedProperty(
            basin3d_vocab=sqlalchemy_opv.basin3d_vocab,
            full_name=sqlalchemy_opv.full_name,
            categories=sqlalchemy_opv.categories.split(","),
            units=sqlalchemy_opv.units
        )

    def _convert_attribute_mapping(self, sqlalchemy_am) -> Optional[basin3d.core.models.AttributeMapping]:
        """
        Convert SQL Alchemy attribute_mapping

        :param sqlalchemy_am: the sqlalchemy attribute mapping
        :return: the basin3d attribute mapping
        """

        attr_type_list = sqlalchemy_am.attr_type.split(MAPPING_DELIMITER)

        basin3d_desc_list = []
        if isinstance(sqlalchemy_am.basin3d_desc, list):
            basin3d_desc_list = sqlalchemy_am.basin3d_desc

        basin3d_desc = []

        for attr_type, desc in zip(attr_type_list, basin3d_desc_list):
            if attr_type == MappedAttributeEnum.OBSERVED_PROPERTY.value:
                op = basin3d.core.models.ObservedProperty(
                    basin3d_vocab=desc.get('basin3d_vocab'),
                    full_name=desc.get('full_name'),
                    categories=desc.get('categories'),
                    units=desc.get('units')
                )
                basin3d_desc.append(op)
            elif attr_type in MappedAttributeEnum.values():
                attr_enum_class = set_mapped_attribute_enum_type(attr_type)
                attr_type_enum = getattr(attr_enum_class, desc)
                basin3d_desc.append(attr_type_enum)
            else:
                basin3d_desc.append(desc)

        return basin3d.core.models.AttributeMapping(
            attr_type=sqlalchemy_am.attr_type,
            basin3d_vocab=sqlalchemy_am.basin3d_vocab,
            basin3d_desc=basin3d_desc,
            datasource_vocab=sqlalchemy_am.datasource_vocab,
            datasource_desc=sqlalchemy_am.datasource_desc,
            datasource=basin3d.core.models.DataSource(id=sqlalchemy_am.datasource.plugin_id,
                                                      name=sqlalchemy_am.datasource.name,
                                                      location=sqlalchemy_am.datasource.location,
                                                      id_prefix=sqlalchemy_am.datasource.id_prefix, )
        )

    def _convert_basin3d_attr_mapping_basin3d_desc(self, basin3d_desc: list) -> list:
        """
        Convert the basin3d_desc to a JSON ready format

        :param basin3d_desc: the basin3d_desc, a list of observed properties, enums, or strings
        :return: a list of JSON ready basin3d_desc
        """
        json_ready_basin3d_desc = []

        for desc in basin3d_desc:
            if isinstance(desc, basin3d.core.models.ObservedProperty):
                json_ready_basin3d_desc.append(desc.to_dict())
            elif isinstance(desc, BaseEnum):
                json_ready_basin3d_desc.append(desc.value)
            else:
                json_ready_basin3d_desc.append(desc)

        return json_ready_basin3d_desc

    def _get_observed_property(self, basin3d_vocab) -> Optional[basin3d.core.models.ObservedProperty]:
        """
        Access a single observed property variable

        :param basin3d_vocab: the observed property name
        :return: the observed property, or None if not found
        """
        session = sqlalchemy_models.Session()

        try:
            opv = session.query(sqlalchemy_models.ObservedProperty).filter_by(basin3d_vocab=basin3d_vocab).one_or_none()
            if opv:
                return self._convert_observed_property(opv)
            return None
        except SQLAlchemyError as e:
            if not isinstance(e, NoResultFound):
                raise e
            return None
        finally:
            session.close()

    def _get_attribute_mapping(self, datasource_id, attr_type, basin3d_vocab, datasource_vocab, **kwargs) -> Optional[
            basin3d.core.models.AttributeMapping]:
        """
        Access a single attribute mapping. If not found, return None.

        :param datasource_id: the datasource identifier, e.g., plugin_id
        :param attr_type: the attribute type, e.g., STATISTIC
        :param basin3d_vocab: the BASIN-3D vocabulary, e.g., Max
        :param datasource_vocab: the datasource vocabulary, e.g., max
        :param kwargs: additional keyword arguments. Not used.

        :return: the attribute mapping, or None if not found
        """
        if not self.is_initialized():
            raise CatalogException("Datasource catalog has not been initialized")

        session = sqlalchemy_models.Session()

        try:
            opv = session.query(sqlalchemy_models.AttributeMapping).filter_by(
                datasource_id=datasource_id, attr_type=attr_type, basin3d_vocab=basin3d_vocab,
                datasource_vocab=datasource_vocab).one_or_none()
            if opv:
                return self._convert_attribute_mapping(opv)
            return None
        except SQLAlchemyError as e:
            if not isinstance(e, NoResultFound):
                raise e
            return None
        finally:
            session.close()

    def find_observed_property(self, basin3d_vocab) -> Optional[basin3d.core.models.ObservedProperty]:
        """
        Return the :class:`basin3d.models.ObservedProperty` object for the BASIN-3D vocabulary specified.

        :param basin3d_vocab: BASIN-3D vocabulary
        :return: a :class:`basin3d.models.ObservedProperty` object
        """
        if not self.is_initialized():
            msg = "Variable Store has not been initialized."
            logger.critical(msg)
            raise CatalogException(msg)

        return self._get_observed_property(basin3d_vocab)

    def find_observed_properties(self, basin3d_vocab: Optional[List[str]] = None) -> Iterator[
            Optional[basin3d.core.models.ObservedProperty]]:
        """
        Report the observed_properties available based on the BASIN-3D vocabularies specified. If no BASIN-3D vocabularies are specified, then return all observed properties available.

        :param basin3d_vocab: list of the BASIN-3D observed properties
        :return: generator that yields :class:`basin3d.models.ObservedProperty` objects
        """
        if not self.is_initialized():
            msg = "Variable Store has not been initialized."
            logger.critical(msg)
            raise CatalogException(msg)

        session = sqlalchemy_models.Session()

        try:
            if not basin3d_vocab:
                for opv in session.query(sqlalchemy_models.ObservedProperty).all():
                    yield self._convert_observed_property(opv)
            else:
                for b3d_vocab in basin3d_vocab:
                    b3d_opv: Optional[basin3d.core.models.ObservedProperty] = self._get_observed_property(b3d_vocab)
                    if b3d_opv is not None:
                        yield b3d_opv
                    else:
                        logger.warning(f'BASIN-3D does not support variable {b3d_vocab}')
        finally:
            session.close()

    def find_datasource_attribute_mapping(self, datasource_id: str, attr_type: str, datasource_vocab: str) -> Optional[
            basin3d.core.models.AttributeMapping]:
        """
        Find the datasource attribute vocabulary to BASIN-3D mapping given a specific datasource_id, attr_type, and datasource attr_vocab.

        :param datasource_id: the datasource identifier
        :param attr_type: attribute type
        :param datasource_vocab: the datasource attribute vocabulary
        :return: a :class:`basin3d.models.AttributeMapping` object

        """
        if not self.is_initialized():
            msg = 'Attribute Store has not been initialized.'
            logger.critical(msg)
            raise CatalogException(msg)

        session = sqlalchemy_models.Session()

        msg = (f'No mapping was found for attr: "{attr_type}" and for datasource vocab: "{datasource_vocab}" '
               f'in datasource: "{datasource_id}".')

        try:
            datasource = session.query(sqlalchemy_models.DataSource).filter_by(plugin_id=datasource_id).one_or_none()
            opv = None
            query_params = {}
            if datasource is None:
                msg = f'No datasource was found for id "{datasource_id}".'
                basin3d_datasource = DATASOURCE_NONE
            else:
                # Set up the search parameters
                query_params = {
                    'datasource_id': datasource.id,
                    'datasource_vocab': datasource_vocab
                }
                basin3d_datasource = \
                    basin3d.core.models.DataSource(
                        id=datasource.plugin_id, name=datasource.name, location=datasource.location, id_prefix=datasource.id_prefix)  # type: ignore

            # Set up empty AttributeMapping in case where mapping is not found or another error occurs
            attr_mapping = basin3d.core.models.AttributeMapping(attr_type=attr_type, basin3d_vocab=NO_MAPPING_TEXT,
                                                                basin3d_desc=[],
                                                                datasource_vocab=datasource_vocab, datasource_desc=msg,
                                                                datasource=basin3d_datasource)
            if datasource:
                opv = session.query(sqlalchemy_models.AttributeMapping).filter_by(**query_params).filter(
                    sqlalchemy_models.AttributeMapping.attr_type.contains(attr_type)).one_or_none()
            if opv is None:
                return attr_mapping
            return self._convert_attribute_mapping(opv)
        except SQLAlchemyError as e:
            if not isinstance(e, NoResultFound):
                raise e
            return attr_mapping
        finally:
            session.close()

    def find_attribute_mappings(self, datasource_id: str = None, attr_type: str = None,
                                attr_vocab: Union[str, List] = None,
                                from_basin3d: bool = False) -> Iterator[basin3d.core.models.AttributeMapping]:
        """
        Find the list of attribute mappings given the specified fields. Exact matches are returned
        (see attr_vocab formats below for BASIN-3D vocab nuances).
        If no fields are specified, all registered attribute mappings will be returned.

        :param datasource_id: the datasource identifier
        :param attr_type: the attribute type
        :param attr_vocab: the attribute vocabulary, the formats are one of the following:
                           1) datasource vocab that is a complete vocab
                           2) BASIN-3D vocab that is a complete vocab for a given attr_type regardless if it is compound or not.
                           3) BASIN-3D vocab that is compound and fully specified for each attr_type either with the complete vocab or with wildcards.
        :param from_basin3d: boolean that says whether the attr_vocab is a BASIN-3D vocabulary. If not, then this a datasource vocabulary.
        :return: generator that yields :class:`basin3d.models.AttributeMapping` objects

        """

        if not self.is_initialized():
            msg = 'Attribute Store has not been initialized.'
            logger.critical(msg)
            raise CatalogException(msg)

        def construct_attr_vocab_query(attr_vocab_list, is_from_basin3d):
            from sqlalchemy import or_
            query = []
            for a_vocab in attr_vocab_list:
                if not is_from_basin3d:
                    query.append(sqlalchemy_models.AttributeMapping.datasource_vocab == a_vocab)
                elif MAPPING_DELIMITER in a_vocab:
                    query.append(sqlalchemy_models.AttributeMapping.basin3d_vocab.op('regexp')(a_vocab))
                else:
                    query.append(or_(
                        sqlalchemy_models.AttributeMapping.basin3d_vocab == a_vocab,
                        sqlalchemy_models.AttributeMapping.basin3d_vocab.op('regexp')(f'.*:{a_vocab}'),
                        sqlalchemy_models.AttributeMapping.basin3d_vocab.op('regexp')(f'{a_vocab}:.*'),
                        sqlalchemy_models.AttributeMapping.basin3d_vocab.op('regexp')(f'.*:{a_vocab}:.*')
                    ))
            return or_(*query)

        session = sqlalchemy_models.Session()
        query_params = []

        if datasource_id is not None:
            try:
                ds = session.query(sqlalchemy_models.DataSource).filter_by(plugin_id=datasource_id).one_or_none()
                if ds is None:
                    logger.warning(
                        f'No datasource for datasource_id {datasource_id} was found. Check plugin initialization')
                    raise CatalogException(f'No datasource for datasource_id {datasource_id} was found.')
                else:
                    query_params.append(sqlalchemy_models.AttributeMapping.datasource_id == ds.id)
            except SQLAlchemyError as e:
                if not isinstance(e, NoResultFound):
                    raise CatalogException(e)

        if attr_type is not None:
            if attr_type not in MappedAttributeEnum.values():
                logger.warning(f'Attribute type {attr_type} is invalid')
                raise CatalogException(f'Attribute type {attr_type} is invalid')
            else:
                query_params.append(sqlalchemy_models.AttributeMapping.attr_type.contains(attr_type))

        if attr_vocab:
            if isinstance(attr_vocab, str):
                attr_vocab = [attr_vocab]
            elif not isinstance(attr_vocab, list):
                raise CatalogException("attr_vocab must be a str or list")
            attr_vocab_query = construct_attr_vocab_query(attr_vocab, from_basin3d)
            query_params.append(attr_vocab_query)

        try:
            attr_mappings = session.query(sqlalchemy_models.AttributeMapping).filter(*query_params).all()
            vocab_source_type = 'datasource' if not from_basin3d else 'BASIN-3D'

            if not attr_mappings:
                logger.info(
                    f'No attribute mappings found for specified parameters: datasource id = '
                    f'"{datasource_id}", attribute type = "{attr_type}", {vocab_source_type} '
                    f'vocabularies: {attr_vocab and ",".join(attr_vocab) or None}.')
            elif attr_vocab and len(attr_mappings) != len(attr_vocab):
                # Find missing vocab in attr_mappings
                if from_basin3d:
                    not_found_attr_vocab = [vocab for vocab in attr_vocab if
                                            vocab not in [am.basin3d_vocab for am in attr_mappings]]
                else:
                    not_found_attr_vocab = [vocab for vocab in attr_vocab if
                                            vocab not in [am.datasource_vocab for am in attr_mappings]]
                logger.warning(
                    f'No attribute mappings found for the following {vocab_source_type} '
                    f'vocabularies: {",".join(not_found_attr_vocab)}. Note: specified datasource id = {datasource_id} and attribute type = {attr_type}.')
        except SQLAlchemyError as e:
            if not isinstance(e, NoResultFound):
                raise e

        for attr_mapping in attr_mappings:
            value = self._convert_attribute_mapping(attr_mapping)
            if value:
                yield value

        session.close()

    def _init_catalog(self, **kwargs):
        """
        Initialize the catalog database
        """
        if not self.is_initialized():
            session = sqlalchemy_models.Session()

            from basin3d.core.plugin import PluginMount
            for name, plugin in PluginMount.plugins.items():

                # Were the plugins passed in? If so, only load the plugins that are in the list
                if ("plugin_ids" in kwargs and plugin.get_meta().id in kwargs[
                   "plugin_ids"]) or "plugin_ids" not in kwargs or kwargs["plugin_ids"] == []:
                    module_name = plugin.__module__
                    class_name = plugin.__name__

                    logger.info("Loading Plugin = {}.{}".format(module_name, class_name))

                    try:
                        datasource = session.query(sqlalchemy_models.DataSource).filter_by(
                            plugin_id=plugin.get_meta().id).one_or_none()
                        if datasource is None:
                            logger.info("Registering NEW Data Source Plugin '{}.{}'".format(module_name, class_name))
                            datasource = sqlalchemy_models.DataSource()
                            if hasattr(plugin.get_meta(), "connection_class"):
                                datasource.credentials = plugin.get_meta().connection_class.get_credentials_format()

                        # Update the datasource
                        datasource.plugin_id = plugin.get_meta().id
                        datasource.name = plugin.get_meta().name
                        datasource.location = plugin.get_meta().location
                        datasource.id_prefix = plugin.get_meta().id_prefix
                        datasource.plugin_module = module_name
                        datasource.plugin_class = class_name
                        session.add(datasource)
                        session.commit()
                        logger.info("Updated Data Source '{}'".format(plugin.get_meta().id))
                    except SQLAlchemyError as e:
                        session.rollback()
                        logger.error(f"Error loading plugin {module_name}.{class_name}: {e}")
                    finally:
                        session.close()

    def _insert(self, record):
        """
        Insert the record. This method is used to insert the record into the database.


        :param record: the record to insert
        """
        if self.is_initialized():
            session = sqlalchemy_models.Session()
            try:
                if isinstance(record, basin3d.core.models.ObservedProperty):
                    try:
                        p = sqlalchemy_models.ObservedProperty(
                            basin3d_vocab=record.basin3d_vocab,
                            full_name=record.full_name,
                            categories=",".join(record.categories),  # type: ignore
                            units=record.units
                        )
                        session.add(p)
                        session.commit()
                        logger.info(f'inserted {record.basin3d_vocab}')
                    except IntegrityError as ie:
                        # This object has already been loaded
                        session.rollback()
                        logger.debug(f'Integrity error for OP: {ie}')
                    except Exception as e:
                        session.rollback()
                        logger.warning(f"Error Registering ObservedProperty '{record.basin3d_vocab}': {str(e)}")

                elif isinstance(record, basin3d.core.models.AttributeMapping):
                    try:
                        ds_name = session.query(sqlalchemy_models.DataSource).filter_by(
                            plugin_id=record.datasource.id).one_or_none()
                        if ds_name is None:
                            raise ValueError(f"DataSource with id {record.datasource.plugin_id} not found")

                        record_basin3d_desc = self._convert_basin3d_attr_mapping_basin3d_desc(record.basin3d_desc)

                        p = sqlalchemy_models.AttributeMapping(
                            datasource=ds_name,
                            attr_type=record.attr_type,
                            basin3d_vocab=record.basin3d_vocab,
                            basin3d_desc=record_basin3d_desc,
                            datasource_vocab=record.datasource_vocab,
                            datasource_desc=record.datasource_desc
                        )
                        session.add(p)
                        session.commit()
                        logger.info(f'inserted {record.datasource_vocab} mapping attribute')
                    except IntegrityError:
                        # This object has already been loaded
                        session.rollback()
                        logger.info(f'Warning: skipping AttributeMapping "{record.basin3d_vocab}". Already loaded.')
                    except Exception as e:
                        session.rollback()
                        logger.info(f'Error Registering AttributeMapping "{record.basin3d_vocab}": {str(e)}')
                elif isinstance(record, basin3d.core.models.DataSource):
                    pass  # Do nothing and don't error out
                else:
                    msg = f"Record type {type(record)} not supported for insertion into a {self.__class__.__name__}"
                    logger.critical(msg)
                    raise CatalogException(msg)
            finally:
                session.close()
        else:
            msg = 'Could not insert record. Catalog not initialized.'
            logger.critical(msg)
            raise CatalogException(msg)



"""

.. currentmodule:: basin3d.core.connection

:platform: Unix, Mac
:synopsis: BASIN-3D ``DataSource`` connection classes
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
import requests
import yaml
from basin3d.core import monitor

from basin3d.core.access import get_url, post_url

logger = monitor.get_logger(__name__)


class HTTPConnectionDataSource(object):
    """
    Class for handling Authentication and authorization of
    :class:`basin3d.models.DataSource` over HTTP


    :param datasource: the datasource to authenticate and authorize via HTTP
    :type datasource: :class:`basin3d.models.DataSource` instance
    """

    def __init__(self, datasource, *args, credentials=None, verify_ssl=False, **kwargs):
        self.datasource = datasource
        self.credentials = credentials
        self.verify_ssl = verify_ssl

    def login(self):
        """
        Login to the :class:`basin3d.models.DataSource`

        :return: JSON response
        :rtype: dict
        """
        raise NotImplementedError

    def logout(self):
        """
        Login out of the :class:`basin3d.models.DataSource`

        :return: None
        """
        raise NotImplementedError

    def get(self, url_part, params=None, headers=None):
        """
        The resources at the spedicfied url

        :param url_part:
        :param params:
        :param headers:
        :return:
        """
        raise NotImplementedError

    def post(self, url_part, params=None, headers=None):
        """
        The resources at the spedicfied url

        :param url_part:
        :param params:
        :param headers:
        :return:
        """
        raise NotImplementedError

    @classmethod
    def get_credentials_format(cls):
        """
        This returnes the format that the credentials are stored in the DB
        :return: The format for the credentials
        """
        raise NotImplementedError


class InvalidOrMissingCredentials(Exception):
    """The credentials are invalid or missing"""
    pass


class HTTPConnectionOAuth2(HTTPConnectionDataSource):
    """
    Class for handling Authentication and authorization of
    :class:`basin3d.models.DataSource` over HTTP with OAuth2


    :param datasource: the datasource to authenticate and authorize via HTTP
    :type datasource: :class:`basin3d.models.DataSource` instance
    :param auth_token_path: The url part for requesting a token
    :param revoke_token_path: The url part for revoking a valid token
    :param auth_scope: The scope of the token being requested (e.g read, write, group)
    :param grant_type: The type of oauth2 grant (e.g client_credentials, password,
            refresh_token, authorization_code)

    """

    CREDENTIALS_FORMAT = 'client_id:\nclient_secret:\n'

    def __init__(self, datasource, *args, auth_token_path="o/token/",
                 revoke_token_path="o/revoke_token/", auth_scope="read",
                 grant_type="client_credentials",
                 **kwargs):

        super(HTTPConnectionOAuth2, self).__init__(datasource, *args, **kwargs)
        self.token = None
        self.auth_token_path = auth_token_path
        self.revoke_token_path = revoke_token_path
        self.auth_scope = auth_scope
        self.grant_type = grant_type
        self.client_id, self.client_secret = self._load_credentials(datasource)

    def _validate_credentials(self):
        """
        Validate the Data Source credentials

        :return: TRUE if the credentials are valid
        :rtype: boolean
        """

        # There should be a client_id and client secret
        return "client_id" in self.credentials.keys() and "client_secret" in self.credentials.keys() \
               and self.credentials["client_id"] and self.credentials["client_secret"]

    def _load_credentials(self, datasource):
        """
        Get the credentials from Data Source. If the
        credentials are invalid `None` is returned.

        :param datasource: The datasource object
        :type datasource: :class:`basin3d.models.DataSource`
        :return: tuple of client_id and client_secret
        :rtype: tuple
        """

        self.credentials = datasource.credentials  # Access the credentials

        # If there are credentials then make the api call
        if self.credentials:
            self.credentials = yaml.load(self.credentials, Loader=yaml.FullLoader)
            if self._validate_credentials():
                return self.credentials["client_id"], self.credentials["client_secret"]

        raise InvalidOrMissingCredentials("client_id and client_secret are missing or invalid")

    def login(self):
        """
        Get a token

        OAuth Client credentials (client_id, client_secret) stored in the
        DataSource.

        - *Url:* `https://<datasource location>/<auth_token_path>`
        - *Scope:* <token_scope>
        - *Grant Type:* <grant_type>
        - *Client Id:* stored in encrypted :class:`basin3d.models.DataSource` field
        - *Client Secret:* stored in encrypted :class:`basin3d.models.DataSource` field

        Example JSON Response::

            {
                "access_token": "<your_access_token>",
                "token_type": "Bearer",
                "expires_in": 36000,
                "refresh_token": "<your_refresh_token>",
                "scope": "read"
            }


        """

        # Build the authentication url
        url = '{}{}'.format(self.datasource.location, self.auth_token_path)
        try:

            # Login to the Data Source
            res = requests.post(url, params={"scope": self.auth_scope, "grant_type": self.grant_type},
                                auth=(self.client_id, self.client_secret),
                                verify=self.verify_ssl)

            # Validate the response
            if res.status_code != requests.codes.ok:
                logger.error("Authentication  error {}: {}".format(url, res.content))
                return None

            # Get the JSON content (This has the token)
            result_json = res.json()
            self.token = result_json
        except Exception as e:
            logger.error("Authentication  error {}: {}".format(url, e))
            # TODO: create exception for this
            # Access is denied!!
            raise Exception("AccessDenied")

    def get(self, url_part, params=None, headers=None):
        """
        Login Data Source if not already logged in.
        Access url with the Authorization header and the access token

        Authorization Header:
            - Authorization": "{token_type} {access_token}

        :param url_part: The url part to request
        :param params: additional parameters for the request
        :type params: dict
        :param headers: request headers
        :return: None
        :raises: PermissionDenied
        """
        self._validate_token()

        # Prepare the Authorization header
        auth_headers = {"Authorization": "{token_type} {access_token}".format(**self.token)}
        if headers:
            auth_headers.update(headers)

        return get_url(url_part, params=params, headers=auth_headers, verify=self.verify_ssl)

    def post(self, url_part, params=None, headers=None):
        """
        Login Data Source if not already logged in.
        Access url with the Authorization header and the access token

        Authorization Header:
            - Authorization": "{token_type} {access_token}

        :param url_part: The url part to request
        :param params: additional parameters for the request
        :type params: dict
        :param headers: request headers
        :return: None
        :raises: PermissionDenied
        """

        self._validate_token()

        # Prepare the Authorization header
        auth_headers = {"Authorization": "{token_type} {access_token}".format(**self.token)}
        if headers:
            auth_headers.update(headers)

        return post_url(url_part, params=params, headers=auth_headers, verify=self.verify_ssl)

    def logout(self):
        """
        Revokes atoken

        :param token: The current Token
        :return: None
        """

        # Build the authentication url for revoking the token
        url = '{}{}'.format(self.datasource.location, self.revoke_token_path)

        # Request the token to be revoked
        if self.token:
            res = requests.post(url, params={"token": self.token["access_token"],
                                             "client_id": self.client_id},
                                auth=(self.client_id, self.client_secret),
                                verify=self.verify_ssl)

            # Validate the success of the token revocation
            if res.status_code != 200:
                logger.warning("Problem encountered revoking token for '{}' HTTP status {} -- {}".format(
                    self.datasource.name,
                    res.status_code, res.content.decode('utf-8')))

    def _validate_token(self):
        """
        Validate the connection token
        :return:
        """
        if not self.token:
            self.login()
        if not self.token:
            # TODO: create exception for this
            # Access is denied!!
            raise Exception("AccessDenied")


class HTTPConnectionTokenAuth(HTTPConnectionDataSource):
    """
    Class for handling Authentication and authorization of
    :class:`basin3d.models.DataSource` over HTTP with Tokens
    :param datasource: the datasource to authenticate and authorize via HTTP
    :type datasource: :class:`basin3d.models.DataSource` instance
    :param login_path: The url part for requesting a token
    """

    CREDENTIALS_FORMAT = 'username: \npassword: \n'

    def __init__(self, datasource, login_path='api/login'):
        """
        Initialize HTTPTokenAuth
        :param datasource:
        :param login_path:
        """

        super(HTTPConnectionTokenAuth, self).__init__(datasource)
        self.userpass = self._load_credentials()
        self.login_path = login_path

    def _load_credentials(self):
        """
        Load the credentials
        :return:
        """
        credentials = self.datasource.credentials  # Access the credentials
        self.credentials = yaml.load(credentials, Loader=yaml.FullLoader)
        # If there are credentials then get the monitoring features
        if self._validate_credentials(self.credentials):
            return self.credentials
        raise InvalidOrMissingCredentials(
            f'Invalid or Missing Credentials - Data Source: {self.datasource.name}')

    @staticmethod
    def _validate_credentials(credentials):
        """
        Validate the credentials
        :param credentials:
        :return:
        """
        return credentials and "username" in credentials.keys() and "password" in credentials.keys() \
            and credentials["username"] and credentials["password"]

    def login(self):
        """
        Get a Token
        :return: JSON response
        :rtype: dict
        """
        try:
            res = requests.get(f'{self.datasource.location}{self.login_path}',
                               params=self.userpass, verify=self.verify_ssl)
            result_json = res.json()
            if result_json[0] == 0:
                return result_json[1]['token']
            elif '{' in result_json[0]:
                return result_json
            else:
                raise InvalidOrMissingCredentials(
                    f'Datasource \'{self.datasource.name}\' Error ({self.login_path}): {result_json}')

        except InvalidOrMissingCredentials as imc:
            raise imc
        except Exception as e:
            raise InvalidOrMissingCredentials(
                f'Datasource \'{self.datasource.name}\' Error ({self.login_path}): {e}')

    def logout(self):
        pass

    def _submit_url(self, url_part, params=None, headers=None, request_method=get_url):
        """
        Interact with the datasource API
        :param url_part:
        :param params:
        :param headers:
        :param request_method:
        :return:
        """

        token_params = [('token', self.login())]
        if params:

            # loop through parameters
            for item in params.items():

                # separate lists into tuples
                if isinstance(item[1], list):
                    for v in item[1]:
                        token_params.append((item[0], v))
                else:
                    token_params.append(item)

        # Check if the url_part contains the datasource location
        url = url_part
        if not url_part.startswith(self.datasource.location):
            url = f'{self.datasource.location}{url_part}'

        return request_method(url, params=token_params,
                              headers=headers, verify=self.verify_ssl)

    def get(self, url_part, params=None, headers=None):
        """
        Get the url
        :param url_part: relative location of the requeste URL
        :param params: The  query parameters
        :type params: dict or list of 2-tuples
        :param headers:
        :return:
        """
        return self._submit_url(url_part, params, headers)

    def post(self, url_part, params=None, headers=None):
        """
        Post to the url
        :param url_part: relative location of the requeste URL
        :param params: The  query parameters
        :type params: dict or list of 2-tuples
        :param headers:
        :return:
        """
        return self._submit_url(url_part, params, headers, post_url)
